from allianceauth import hooks
from django.utils.translation import gettext_lazy as _

from ..menu.hooks import MenuItemHook
from .hooks import ServicesHook


class Services(MenuItemHook):
    def __init__(self):
        MenuItemHook.__init__(
            self,
            _('Services'),
            'fa-solid fa-gears',
            'services:services', 100
        )

    def render(self, request):
        for svc in ServicesHook.get_services():
            if svc.service_active_for_user(request.user):
                return MenuItemHook.render(self, request)
        return ''


@hooks.register('menu_item_hook')
def register_menu():
    return Services()



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class Ips4ServiceConfig(AppConfig):
    name = 'allianceauth.services.modules.ips4'
    label = 'ips4'
    verbose_name = _('IPS4 Service')



"""
Example
=======

.. code-block:: python

    from allianceauth.notifications.models import Notification


    def notify_user_view(request):
        '''Simple view sending a notification to the user'''

        Notification.objects.notify_user(
                user=request.user,
                title="Some title",
                message="Some message",
                level=Notification.Level.INFO,
            )

"""
from .core import notify  # noqa: F401



from unittest.mock import patch

from django.db.models import QuerySet
from django.test import TestCase

from allianceauth.menu.constants import MenuItemType
from allianceauth.menu.models import MenuItem

from .factories import (
    create_app_menu_item,
    create_folder_menu_item,
    create_link_menu_item,
    create_menu_item_from_hook,
    create_menu_item_hook_function,
)
from .utils import PACKAGE_PATH


class TestMenuItemQuerySet(TestCase):
    def test_should_add_item_type_field(self):
        # given
        app_item = create_app_menu_item()
        link_item = create_link_menu_item()
        folder_item = create_folder_menu_item()

        # when
        result: QuerySet[MenuItem] = MenuItem.objects.annotate_item_type_2()

        # then
        for obj in [app_item, link_item, folder_item]:
            obj = result.get(pk=app_item.pk)
            self.assertEqual(obj.item_type_2, obj.item_type)

    def test_should_filter_folders(self):
        # given
        create_app_menu_item()
        create_link_menu_item()
        folder_item = create_folder_menu_item()

        # when
        result: QuerySet[MenuItem] = MenuItem.objects.filter_folders()

        # then
        item_pks = set(result.values_list("pk", flat=True))
        self.assertSetEqual(item_pks, {folder_item.pk})


@patch(PACKAGE_PATH + ".managers.get_hooks", spec=True)
class TestMenuItemManagerSyncAll(TestCase):
    def test_should_create_new_items_from_hooks_when_they_do_not_exist(
        self, mock_get_hooks
    ):
        # given
        mock_get_hooks.return_value = [create_menu_item_hook_function(text="Alpha")]

        # when
        MenuItem.objects.sync_all()

        # then
        self.assertEqual(MenuItem.objects.count(), 1)
        obj = MenuItem.objects.first()
        self.assertEqual(obj.item_type, MenuItemType.APP)
        self.assertEqual(obj.text, "Alpha")

    def test_should_update_existing_app_items_when_changed_only(self, mock_get_hooks):
        # given
        menu_hook_1 = create_menu_item_hook_function(text="Alpha", order=1)
        menu_hook_2 = create_menu_item_hook_function(text="Bravo", order=2)
        mock_get_hooks.return_value = [menu_hook_1, menu_hook_2]
        create_menu_item_from_hook(menu_hook_1, text="name has changed", order=99)
        create_menu_item_from_hook(menu_hook_2)

        # when
        MenuItem.objects.sync_all()

        # then
        self.assertEqual(MenuItem.objects.count(), 2)

        obj = MenuItem.objects.get(text="Alpha")
        self.assertEqual(obj.item_type, MenuItemType.APP)
        self.assertEqual(obj.order, 99)

        obj = MenuItem.objects.get(text="Bravo")
        self.assertEqual(obj.item_type, MenuItemType.APP)
        self.assertEqual(obj.order, 2)

    def test_should_remove_obsolete_app_items_but_keep_user_items(self, mock_get_hooks):
        # given
        menu_hook = create_menu_item_hook_function(text="Alpha")
        mock_get_hooks.return_value = [menu_hook]
        create_app_menu_item(text="Bravo")  # obsolete item
        link_item = create_link_menu_item()
        folder_item = create_folder_menu_item()

        # when
        MenuItem.objects.sync_all()

        # then
        self.assertEqual(MenuItem.objects.count(), 3)
        obj = MenuItem.objects.get(text="Alpha")
        self.assertTrue(obj.item_type, MenuItemType.APP)
        self.assertIn(link_item, MenuItem.objects.all())
        self.assertIn(folder_item, MenuItem.objects.all())



from django.urls import path
from . import views

app_name = 'corputils'
urlpatterns = [
    path('', views.corpstats_view, name='view'),
    path('add/', views.corpstats_add, name='add'),
    path('<int:corp_id>/', views.corpstats_view, name='view_corp'),
    path('<int:corp_id>/update/', views.corpstats_update, name='update'),
    path('search/', views.corpstats_search, name='search'),
]



from django.test import TestCase

from allianceauth.menu.constants import MenuItemType

from .factories import (
    create_app_menu_item,
    create_folder_menu_item,
    create_link_menu_item,
)


class TestMenuItem(TestCase):
    def test_str(self):
        # given
        obj = create_link_menu_item()
        # when
        result = str(obj)
        # then
        self.assertIsInstance(result, str)

    def test_should_return_item_type(self):
        # given
        app_item = create_app_menu_item()
        link_item = create_link_menu_item()
        folder_item = create_folder_menu_item()

        cases = [
            (app_item, MenuItemType.APP),
            (link_item, MenuItemType.LINK),
            (folder_item, MenuItemType.FOLDER),
        ]
        # when
        for obj, expected in cases:
            with self.subTest(type=expected):
                self.assertEqual(obj.item_type, expected)

    def test_should_identify_if_item_is_a_child(self):
        # given
        folder = create_folder_menu_item()
        child = create_link_menu_item(parent=folder)
        not_child = create_link_menu_item()

        cases = [(child, True), (not_child, False)]
        # when
        for obj, expected in cases:
            with self.subTest(type=expected):
                self.assertIs(obj.is_child, expected)

    def test_should_identify_if_item_is_a_folder(self):
        # given
        app_item = create_app_menu_item()
        link_item = create_link_menu_item()
        folder_item = create_folder_menu_item()

        cases = [
            (app_item, False),
            (link_item, False),
            (folder_item, True),
        ]
        # when
        for obj, expected in cases:
            with self.subTest(type=expected):
                self.assertIs(obj.is_folder, expected)

    def test_should_identify_if_item_is_user_defined(self):
        # given
        app_item = create_app_menu_item()
        link_item = create_link_menu_item()
        folder_item = create_folder_menu_item()

        cases = [
            (app_item, False),
            (link_item, True),
            (folder_item, True),
        ]
        # when
        for obj, expected in cases:
            with self.subTest(type=expected):
                self.assertIs(obj.is_user_defined, expected)

    def test_should_identify_if_item_is_an_app_item(self):
        # given
        app_item = create_app_menu_item()
        link_item = create_link_menu_item()
        folder_item = create_folder_menu_item()

        cases = [
            (app_item, True),
            (link_item, False),
            (folder_item, False),
        ]
        # when
        for obj, expected in cases:
            with self.subTest(type=expected):
                self.assertIs(obj.is_app_item, expected)

    def test_should_identify_if_item_is_a_link_item(self):
        # given
        app_item = create_app_menu_item()
        link_item = create_link_menu_item()
        folder_item = create_folder_menu_item()

        cases = [
            (app_item, False),
            (link_item, True),
            (folder_item, False),
        ]
        # when
        for obj, expected in cases:
            with self.subTest(type=expected):
                self.assertIs(obj.is_link_item, expected)

    def test_should_not_allow_creating_invalid_app_item(self):
        # when
        obj = create_app_menu_item(hook_hash="")

        # then
        obj.refresh_from_db()
        self.assertIsNone(obj.hook_hash)


class TestMenuItemToHookObj(TestCase):
    def test_should_create_from_link_item(self):
        # given
        obj = create_link_menu_item(text="Alpha")

        # when
        hook_obj = obj.to_hook_obj()

        # then
        self.assertEqual(hook_obj.text, "Alpha")
        self.assertEqual(hook_obj.url, obj.url)
        self.assertEqual(hook_obj.html_id, "")
        self.assertFalse(hook_obj.is_folder)

    def test_should_create_from_folder(self):
        # given
        obj = create_folder_menu_item(text="Alpha", classes="dummy")

        # when
        hook_obj = obj.to_hook_obj()

        # then
        self.assertEqual(hook_obj.text, "Alpha")
        self.assertEqual(hook_obj.classes, "dummy")
        self.assertEqual(hook_obj.url, "")
        self.assertTrue(hook_obj.html_id)
        self.assertTrue(hook_obj.is_folder)

    def test_should_create_from_folder_and_use_default_icon_classes(self):
        # given
        obj = create_folder_menu_item(classes="")

        # when
        hook_obj = obj.to_hook_obj()

        # then
        self.assertEqual(hook_obj.classes, "fa-solid fa-folder")

    def test_should_create_from_app_item(self):
        # given
        obj = create_app_menu_item(text="Alpha")

        # when
        with self.assertRaises(ValueError):
            obj.to_hook_obj()



# Generated by Django 4.2.16 on 2025-03-26 20:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('timerboard', '0007_alter_timer_structure'),
    ]

    operations = [
        migrations.AlterField(
            model_name='timer',
            name='timer_type',
            field=models.CharField(choices=[('UNSPECIFIED', 'Not Specified'), ('SHIELD', 'Shield'), ('ARMOR', 'Armor'), ('HULL', 'Hull'), ('FINAL', 'Final'), ('ANCHORING', 'Anchoring'), ('UNANCHORING', 'Unanchoring'), ('ABANDONED', 'Abandoned'), ('THEFT', 'Theft')], default='UNSPECIFIED', max_length=254),
        ),
    ]



"""
Migration to AA Framework API method
"""

from django.conf import settings
from django.db import migrations, models

import allianceauth.framework.api.user


class Migration(migrations.Migration):

    dependencies = [
        ("fleetactivitytracking", "0006_auto_20180803_0430"),
    ]

    operations = [
        migrations.AlterField(
            model_name="fatlink",
            name="creator",
            field=models.ForeignKey(
                on_delete=models.SET(allianceauth.framework.api.user.get_sentinel_user),
                to=settings.AUTH_USER_MODEL
            ),
        ),
    ]



from django.test import RequestFactory, TestCase

from allianceauth import views

from .auth_utils import AuthUtils


class TestCustomErrorHandlerViews(TestCase):
    @classmethod
    def setUpTestData(cls) -> None:
        cls.user = AuthUtils.create_user("my_user")
        cls.factory = RequestFactory()

    def test_should_return_status_code_400(self):
        # give
        request = self.factory.get("/")
        request.user = self.user
        # when
        response = views.Generic400Redirect(request)
        # then
        self.assertEqual(response.status_code, 400)

    def test_should_return_status_code_403(self):
        # give
        request = self.factory.get("/")
        request.user = self.user
        # when
        response = views.Generic403Redirect(request)
        # then
        self.assertEqual(response.status_code, 403)

    def test_should_return_status_code_404(self):
        # give
        request = self.factory.get("/")
        request.user = self.user
        # when
        response = views.Generic404Redirect(request)
        # then
        self.assertEqual(response.status_code, 404)

    def test_should_return_status_code_500(self):
        # give
        request = self.factory.get("/")
        request.user = self.user
        # when
        response = views.Generic500Redirect(request)
        # then
        self.assertEqual(response.status_code, 500)



# Generated by Django 4.2.9 on 2024-02-15 00:01

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="MenuItem",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "text",
                    models.CharField(
                        db_index=True,
                        help_text="Text to show on menu",
                        max_length=150,
                        verbose_name="text",
                    ),
                ),
                (
                    "order",
                    models.IntegerField(
                        db_index=True,
                        default=9999,
                        help_text="Order of the menu. Lowest First",
                        verbose_name="order",
                    ),
                ),
                (
                    "is_hidden",
                    models.BooleanField(
                        default=False,
                        help_text="Hide this menu item.If this item is a folder all items under it will be hidden too",
                        verbose_name="is hidden",
                    ),
                ),
                (
                    "hook_hash",
                    models.CharField(
                        default=None,
                        editable=False,
                        max_length=64,
                        null=True,
                        unique=True,
                    ),
                ),
                (
                    "classes",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="Font Awesome classes to show as icon on menu, e.g. <code>fa-solid fa-house</code>",
                        max_length=150,
                        verbose_name="icon classes",
                    ),
                ),
                (
                    "url",
                    models.TextField(
                        default="",
                        help_text="External URL this menu items will link to",
                        verbose_name="url",
                    ),
                ),
                (
                    "parent",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        help_text="Folder this item is in (optional)",
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="children",
                        to="menu.menuitem",
                        verbose_name="folder",
                    ),
                ),
            ],
        ),
    ]



from django.contrib import admin
from django.contrib.auth.models import Group
from .models import AuthTS, Teamspeak3User, StateGroup, TSgroup
from ...admin import ServicesUserAdmin
from allianceauth.groupmanagement.models import ReservedGroupName


@admin.register(Teamspeak3User)
class Teamspeak3UserAdmin(ServicesUserAdmin):
    list_display = ServicesUserAdmin.list_display + (
        'uid',
        'perm_key'
    )
    search_fields = ServicesUserAdmin.search_fields + ('uid', )


@admin.register(AuthTS)
class AuthTSgroupAdmin(admin.ModelAdmin):
    change_list_template = 'admin/teamspeak3/authts/change_list.html'
    ordering = ('auth_group__name', )
    list_select_related = True

    list_display = ('auth_group', '_ts_group')
    list_filter = ('ts_group', )

    fields = ('auth_group', 'ts_group')
    filter_horizontal = ('ts_group',)

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        if db_field.name == 'auth_group':
            kwargs['queryset'] = Group.objects.exclude(name__in=ReservedGroupName.objects.values_list('name', flat=True))
        return super().formfield_for_foreignkey(db_field, request, **kwargs)

    def formfield_for_manytomany(self, db_field, request, **kwargs):
        if db_field.name == 'ts_group':
            kwargs['queryset'] = TSgroup.objects.exclude(ts_group_name__in=ReservedGroupName.objects.values_list('name', flat=True))
        return super().formfield_for_manytomany(db_field, request, **kwargs)

    @admin.display(
        description='ts groups'
    )
    def _ts_group(self, obj):
        return [x for x in obj.ts_group.all().order_by('ts_group_id')]

    # _ts_group.admin_order_field = 'profile__state'


@admin.register(StateGroup)
class StateGroupAdmin(admin.ModelAdmin):
    list_display = ('state', 'ts_group')
    search_fields = ('state__name', 'ts_group__ts_group_name')



import logging

from django.contrib.auth.backends import ModelBackend
from django.contrib.auth.models import User, Permission

from .models import UserProfile, CharacterOwnership, OwnershipRecord


logger = logging.getLogger(__name__)


class StateBackend(ModelBackend):
    @staticmethod
    def _get_state_permissions(user_obj):
        """returns permissions for state of given user object"""
        if hasattr(user_obj, "profile") and user_obj.profile:
            return Permission.objects.filter(state=user_obj.profile.state)
        else:
            return Permission.objects.none()

    def get_state_permissions(self, user_obj, obj=None):
        return self._get_permissions(user_obj, obj, 'state')

    def get_all_permissions(self, user_obj, obj=None):
        if not user_obj.is_active or user_obj.is_anonymous or obj is not None:
            return set()
        if not hasattr(user_obj, '_perm_cache'):
            user_obj._perm_cache = self.get_user_permissions(user_obj)
            user_obj._perm_cache.update(self.get_group_permissions(user_obj))
            user_obj._perm_cache.update(self.get_state_permissions(user_obj))
        return user_obj._perm_cache

    def authenticate(self, request=None, token=None, **credentials):
        if not token:
            return None
        try:
            ownership = CharacterOwnership.objects.get(character__character_id=token.character_id)
            if ownership.owner_hash == token.character_owner_hash:
                logger.debug(f'Authenticating {ownership.user} by ownership of character {token.character_name}')
                if ownership.user.profile.main_character:
                    if ownership.user.profile.main_character.character_id == token.character_id:
                        return ownership.user
                    else:  # this is an alt, enforce main only.
                        return None
            else:
                logger.debug(f'{token.character_name} has changed ownership. Creating new user account.')
                ownership.delete()
                return self.create_user(token)
        except CharacterOwnership.DoesNotExist:
            try:
                # insecure legacy main check for pre-sso registration auth installs
                profile = UserProfile.objects.get(main_character__character_id=token.character_id)
                logger.debug(f'Authenticating {profile.user} by their main character {profile.main_character} without active ownership.')
                # attach an ownership
                token.user = profile.user
                CharacterOwnership.objects.create_by_token(token)
                return profile.user
            except UserProfile.DoesNotExist:
                # now we check historical records to see if this is a returning user
                records = OwnershipRecord.objects.filter(owner_hash=token.character_owner_hash).filter(character__character_id=token.character_id)
                if records.exists():
                    # we've seen this character owner before. Re-attach to their old user account
                    user = records[0].user
                    if user.profile.main_character:
                        if user.profile.main_character.character_id != token.character_id:
                            # this is an alt, enforce main only due to trust issues in SSO.
                            return None

                    token.user = user
                    co = CharacterOwnership.objects.create_by_token(token)
                    logger.debug(f'Authenticating {user} by matching owner hash record of character {co.character}')

                    # set this as their main by default as they have none
                    user.profile.main_character = co.character
                    user.profile.save()
                    return user
            logger.debug(f'Unable to authenticate character {token.character_name}. Creating new user.')
            return self.create_user(token)

    def create_user(self, token):
        username = self.iterate_username(token.character_name)  # build unique username off character name
        user = User.objects.create_user(username, is_active=False)  # prevent login until email set
        user.set_unusable_password()  # prevent login via password
        user.save()
        token.user = user
        co = CharacterOwnership.objects.create_by_token(token)  # assign ownership to this user
        user.profile.main_character = co.character  # assign main character as token character
        user.profile.save()
        logger.debug(f'Created new user {user}')
        return user

    @staticmethod
    def iterate_username(name):
        name = str.replace(name, "'", "")
        name = str.replace(name, ' ', '_')
        if User.objects.filter(username__startswith=name).exists():
            u = User.objects.filter(username__startswith=name)
            num = len(u)
            username = f"{name}_{num}"
            while u.filter(username=username).exists():
                num += 1
                username = f"{name}_{num}"
        else:
            username = name
        return username



"""
Django app configuration for custom_css
"""

# Django
from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class CustomCSSConfig(AppConfig):
    name = "allianceauth.custom_css"
    label = "custom_css"
    verbose_name = _("Custom CSS")



"""
Test sentinel user
"""

import re

# Django
from django.contrib.auth.models import User
from django.test import TestCase

# Alliance Auth
from allianceauth.framework.api.user import (
    get_sentinel_user,
    get_main_character_from_user,
    get_main_character_name_from_user
)
from allianceauth.tests.auth_utils import AuthUtils


class TestSentinelUser(TestCase):
    """
    Tests for the sentinel user
    """

    def test_should_create_user_when_it_does_not_exist(self) -> None:
        """
        Test should create a sentinel user when it doesn't exist

        :return:
        :rtype:
        """

        # when
        user = get_sentinel_user()

        # then
        self.assertEqual(first=user.username, second="deleted")

    def test_should_return_user_when_it_does(self) -> None:
        """
        Test should return sentinel user when it exists

        :return:
        :rtype:
        """

        # given
        User.objects.create_user(username="deleted")

        # when
        user = get_sentinel_user()

        # then
        self.assertEqual(first=user.username, second="deleted")


class TestGetMainForUser(TestCase):
    """
    Tests for get_main_character_from_user
    """

    @classmethod
    def setUpClass(cls) -> None:
        """
        Set up groups and users
        """

        super().setUpClass()

        cls.character_name = "William T. Riker"
        cls.character_name_2 = "Christopher Pike"

        cls.username = re.sub(pattern=r"[^\w\d@\.\+-]", repl="_", string=cls.character_name)
        cls.username_2 = re.sub(
            pattern=r"[^\w\d@\.\+-]", repl="_", string=cls.character_name_2
        )

        cls.user = AuthUtils.create_user(username=cls.username)
        cls.user_without_main = AuthUtils.create_user(
            username=cls.username_2, disconnect_signals=True
        )

        cls.character = AuthUtils.add_main_character_2(
            user=cls.user, name=cls.character_name, character_id=1001
        )


    def test_get_main_character_from_user_should_return_character_name(self):
        """
        Test should return the main character name for a regular user

        :return:
        :rtype:
        """

        character = get_main_character_from_user(user=self.user)

        self.assertEqual(first=character, second=self.character)


    def test_get_main_character_from_user_should_return_none_for_no_main_character(self):
        """
        Test should return None for User without a main character

        :return:
        :rtype:
        """

        character = get_main_character_from_user(user=self.user_without_main)

        self.assertIsNone(obj=character)


    def test_get_main_character_from_user_should_none(self):
        """
        Test should return None when user is None

        :return:
        :rtype:
        """

        user = None

        character = get_main_character_from_user(user=user)

        self.assertIsNone(obj=character)


    def test_get_main_character_name_from_user_should_return_character_name(self):
        """
        Test should return the main character name for a regular user

        :return:
        :rtype:
        """

        character_name = get_main_character_name_from_user(user=self.user)

        self.assertEqual(first=character_name, second=self.character_name)

    def test_get_main_character_name_from_user_should_return_user_name(self):
        """
        Test should return just the username for a user without a main character

        :return:
        :rtype:
        """

        character_name = get_main_character_name_from_user(user=self.user_without_main)

        self.assertEqual(first=character_name, second=self.username_2)

    def test_get_main_character_name_from_user_should_return_sentinel_user(self):
        """
        Test should return "deleted" as username (Sentinel User)

        :return:
        :rtype:
        """

        user = get_sentinel_user()

        character_name = get_main_character_name_from_user(user=user)

        self.assertEqual(first=character_name, second="deleted")

    def test_get_main_character_name_from_user_should_return_sentinel_user_for_none(self):
        """
        Test should return "deleted" (Sentinel User) if user is None

        :return:
        :rtype:
        """

        user = None

        character_name = get_main_character_name_from_user(user=user)

        self.assertEqual(first=character_name, second="deleted")



from django.contrib import admin
from django.contrib.auth.admin import UserAdmin as BaseUserAdmin
from django.contrib.auth.models import Group
from django.contrib.auth.models import Permission as BasePermission
from django.contrib.auth.models import User as BaseUser
from django.db.models import Count, Q
from django.db.models.functions import Lower
from django.db.models.signals import (
    m2m_changed,
    post_delete,
    post_save,
    pre_delete,
    pre_save
)
from django.dispatch import receiver
from django.urls import reverse
from django.utils.html import format_html
from django.utils.text import slugify

from allianceauth.authentication.models import (
    CharacterOwnership,
    OwnershipRecord,
    State,
    UserProfile,
    get_guest_state
)
from allianceauth.eveonline.models import (
    EveAllianceInfo,
    EveCharacter,
    EveCorporationInfo,
    EveFactionInfo
)
from allianceauth.eveonline.tasks import update_character
from allianceauth.hooks import get_hooks
from allianceauth.services.hooks import ServicesHook

from .app_settings import (
    AUTHENTICATION_ADMIN_USERS_MAX_CHARS,
    AUTHENTICATION_ADMIN_USERS_MAX_GROUPS
)
from .forms import UserChangeForm, UserProfileForm


def make_service_hooks_update_groups_action(service):
    """
    Make a admin action for the given service
    :param service: services.hooks.ServicesHook
    :return: fn to update services groups for the selected users
    """
    def update_service_groups(modeladmin, request, queryset):
        if hasattr(service, 'update_groups_bulk'):
            service.update_groups_bulk(queryset)
        else:
            for user in queryset:  # queryset filtering doesn't work here?
                service.update_groups(user)

    update_service_groups.__name__ = str(f'update_{slugify(service.name)}_groups')
    update_service_groups.short_description = f"Sync groups for selected {service.title} accounts"
    return update_service_groups


def make_service_hooks_sync_nickname_action(service):
    """
    Make a sync_nickname admin action for the given service
    :param service: services.hooks.ServicesHook
    :return: fn to sync nickname for the selected users
    """
    def sync_nickname(modeladmin, request, queryset):
        if hasattr(service, 'sync_nicknames_bulk'):
            service.sync_nicknames_bulk(queryset)
        else:
            for user in queryset:  # queryset filtering doesn't work here?
                service.sync_nickname(user)

    sync_nickname.__name__ = str(f'sync_{slugify(service.name)}_nickname')
    sync_nickname.short_description = f"Sync nicknames for selected {service.title} accounts"
    return sync_nickname


class UserProfileInline(admin.StackedInline):
    model = UserProfile
    readonly_fields = ('state',)
    form = UserProfileForm
    verbose_name = ''
    verbose_name_plural = 'Profile'

    def get_formset(self, request, obj=None, **kwargs):
        # main_character field can only show current value or unclaimed alts
        # if superuser, allow selecting from any unclaimed main
        query = Q()
        if obj and obj.profile.main_character:
            query |= Q(pk=obj.profile.main_character_id)
            if request.user.is_superuser:
                query |= Q(userprofile__isnull=True)
            else:
                query |= Q(character_ownership__user=obj)
        formset = super().get_formset(request, obj=obj, **kwargs)

        def get_kwargs(self, index):
            return {'querysets': {'main_character': EveCharacter.objects.filter(query)}}
        formset.get_form_kwargs = get_kwargs
        return formset

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


@admin.display(description="")
def user_profile_pic(obj):
    """profile pic column data for user objects

    works for both User objects and objects with `user` as FK to User
    To be used for all user based admin lists (requires CSS)
    """
    user_obj = obj.user if hasattr(obj, 'user') else obj
    if user_obj.profile.main_character:
        return format_html(
            '<img src="{}" class="img-circle">',
            user_obj.profile.main_character.portrait_url(size=32)
        )
    return None


@admin.display(description="user / main", ordering="username")
def user_username(obj):
    """user column data for user objects

    works for both User objects and objects with `user` as FK to User
    To be used for all user based admin lists
    """
    link = reverse(
        'admin:{}_{}_change'.format(
            obj._meta.app_label,
            type(obj).__name__.lower()
        ),
        args=(obj.pk,)
    )
    user_obj = obj.user if hasattr(obj, 'user') else obj
    if user_obj.profile.main_character:
        return format_html(
            '<strong><a href="{}">{}</a></strong><br>{}',
            link,
            user_obj.username,
            user_obj.profile.main_character.character_name
        )
    return format_html(
        '<strong><a href="{}">{}</a></strong>',
        link,
        user_obj.username,
    )


@admin.display(
    description="Corporation / Alliance (Main)",
    ordering="profile__main_character__corporation_name"
)
def user_main_organization(obj):
    """main organization column data for user objects

    works for both User objects and objects with `user` as FK to User
    To be used for all user based admin lists
    """
    user_obj = obj.user if hasattr(obj, 'user') else obj
    if not user_obj.profile.main_character:
        return ''
    result = user_obj.profile.main_character.corporation_name
    if user_obj.profile.main_character.alliance_id:
        result += f'<br>{user_obj.profile.main_character.alliance_name}'
    elif user_obj.profile.main_character.faction_name:
        result += f'<br>{user_obj.profile.main_character.faction_name}'
    return format_html(result)


class MainCorporationsFilter(admin.SimpleListFilter):
    """Custom filter to filter on corporations from mains only

    works for both User objects and objects with `user` as FK to User
    To be used for all user based admin lists
    """
    title = 'corporation'
    parameter_name = 'main_corporation_id__exact'

    def lookups(self, request, model_admin):
        qs = EveCharacter.objects\
            .exclude(userprofile=None)\
            .values('corporation_id', 'corporation_name')\
            .distinct()\
            .order_by(Lower('corporation_name'))
        return tuple(
            (x['corporation_id'], x['corporation_name']) for x in qs
        )

    def queryset(self, request, qs):
        if self.value() is None:
            return qs.all()
        if qs.model == User:
            return qs.filter(
                profile__main_character__corporation_id=self.value()
            )
        return qs.filter(
            user__profile__main_character__corporation_id=self.value()
        )


class MainAllianceFilter(admin.SimpleListFilter):
    """Custom filter to filter on alliances from mains only

    works for both User objects and objects with `user` as FK to User
    To be used for all user based admin lists
    """
    title = 'alliance'
    parameter_name = 'main_alliance_id__exact'

    def lookups(self, request, model_admin):
        qs = (
            EveCharacter.objects
            .exclude(alliance_id=None)
            .exclude(userprofile=None)
            .values('alliance_id', 'alliance_name')
            .distinct()
            .order_by(Lower('alliance_name'))
        )
        return tuple(
            (x['alliance_id'], x['alliance_name']) for x in qs
        )

    def queryset(self, request, qs):
        if self.value() is None:
            return qs.all()
        if qs.model == User:
            return qs.filter(profile__main_character__alliance_id=self.value())
        return qs.filter(
            user__profile__main_character__alliance_id=self.value()
        )


class MainFactionFilter(admin.SimpleListFilter):
    """Custom filter to filter on factions from mains only

    works for both User objects and objects with `user` as FK to User
    To be used for all user based admin lists
    """
    title = 'faction'
    parameter_name = 'main_faction_id__exact'

    def lookups(self, request, model_admin):
        qs = (
            EveCharacter.objects
            .exclude(faction_id=None)
            .exclude(userprofile=None)
            .values('faction_id', 'faction_name')
            .distinct()
            .order_by(Lower('faction_name'))
        )
        return tuple(
            (x['faction_id'], x['faction_name']) for x in qs
        )

    def queryset(self, request, qs):
        if self.value() is None:
            return qs.all()
        if qs.model == User:
            return qs.filter(profile__main_character__faction_id=self.value())
        return qs.filter(
            user__profile__main_character__faction_id=self.value()
        )


@admin.display(description="Update main character model from ESI")
def update_main_character_model(modeladmin, request, queryset):
    tasks_count = 0
    for obj in queryset:
        if obj.profile.main_character:
            update_character.delay(obj.profile.main_character.character_id)
            tasks_count += 1

    modeladmin.message_user(
        request, f'Update from ESI started for {tasks_count} characters'
    )


class UserAdmin(BaseUserAdmin):
    """Extending Django's UserAdmin model

    Behavior of groups and characters columns can be configured via settings
    """

    inlines = [UserProfileInline]
    ordering = ('username', )
    list_select_related = ('profile__state', 'profile__main_character')
    show_full_result_count = True
    list_display = (
        user_profile_pic,
        user_username,
        '_state',
        '_groups',
        user_main_organization,
        '_characters',
        'is_active',
        'date_joined',
        '_role'
    )
    list_display_links = None
    list_filter = (
        'profile__state',
        'groups',
        MainCorporationsFilter,
        MainAllianceFilter,
        MainFactionFilter,
        'is_active',
        'date_joined',
        'is_staff',
        'is_superuser'
    )
    search_fields = ('username', 'character_ownerships__character__character_name')
    readonly_fields = ('date_joined', 'last_login')
    filter_horizontal = ('groups', 'user_permissions',)
    form = UserChangeForm

    class Media:
        css = {
            "all": ("allianceauth/authentication/css/admin.css",)
        }

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.prefetch_related("character_ownerships__character", "groups")

    def get_form(self, request, obj=None, **kwargs):
        """Inject current request into change form object."""

        MyForm = super().get_form(request, obj, **kwargs)
        if obj:
            class MyFormInjected(MyForm):
                def __new__(cls, *args, **kwargs):
                    kwargs['request'] = request
                    return MyForm(*args, **kwargs)

            return MyFormInjected
        return MyForm

    def get_actions(self, request):
        actions = super().get_actions(request)
        actions[update_main_character_model.__name__] = (
            update_main_character_model,
            update_main_character_model.__name__,
            update_main_character_model.short_description
        )

        for hook in get_hooks('services_hook'):
            svc = hook()
            # Check update_groups is redefined/overloaded
            if svc.update_groups.__module__ != ServicesHook.update_groups.__module__:
                action = make_service_hooks_update_groups_action(svc)
                actions[action.__name__] = (
                    action,
                    action.__name__,
                    action.short_description
                )

            # Create sync nickname action if service implements it
            if svc.sync_nickname.__module__ != ServicesHook.sync_nickname.__module__:
                action = make_service_hooks_sync_nickname_action(svc)
                actions[action.__name__] = (
                    action, action.__name__,
                    action.short_description
                )
        return actions

    def _list_2_html_w_tooltips(self, my_items: list, max_items: int) -> str:
        """converts list of strings into HTML with cutoff and tooltip"""
        items_truncated_str = ', '.join(my_items[:max_items])
        if not my_items:
            result = None
        elif len(my_items) <= max_items:
            result = items_truncated_str
        else:
            items_truncated_str += ', (...)'
            items_all_str = ', '.join(my_items)
            result = format_html(
                '<span data-tooltip="{}" class="tooltip">{}</span>',
                items_all_str,
                items_truncated_str
            )
        return result

    def _characters(self, obj):
        character_ownerships = list(obj.character_ownerships.all())
        characters = [obj.character.character_name for obj in character_ownerships]
        return self._list_2_html_w_tooltips(
            sorted(characters),
            AUTHENTICATION_ADMIN_USERS_MAX_CHARS
        )

    @admin.display(ordering="profile__state")
    def _state(self, obj):
        return obj.profile.state.name

    def _groups(self, obj):
        my_groups = sorted(group.name for group in list(obj.groups.all()))
        return self._list_2_html_w_tooltips(
            my_groups, AUTHENTICATION_ADMIN_USERS_MAX_GROUPS
        )

    def _role(self, obj):
        if obj.is_superuser:
            role = 'Superuser'
        elif obj.is_staff:
            role = 'Staff'
        else:
            role = 'User'
        return role

    def has_change_permission(self, request, obj=None):
        return request.user.has_perm('auth.change_user')

    def has_add_permission(self, request, obj=None):
        return request.user.has_perm('auth.add_user')

    def has_delete_permission(self, request, obj=None):
        return request.user.has_perm('auth.delete_user')

    def get_object(self, *args , **kwargs):
        obj = super().get_object(*args , **kwargs)
        self.obj = obj  # storing current object for use in formfield_for_manytomany
        return obj

    def formfield_for_manytomany(self, db_field, request, **kwargs):
        if db_field.name == "groups":
            groups_qs = Group.objects.filter(authgroup__states__isnull=True)
            obj_state = self.obj.profile.state
            if obj_state:
                matching_groups_qs = Group.objects.filter(authgroup__states=obj_state)
                groups_qs = groups_qs | matching_groups_qs
            kwargs["queryset"] = groups_qs.order_by(Lower("name"))
        return super().formfield_for_manytomany(db_field, request, **kwargs)

    def get_readonly_fields(self, request, obj=None):
        if obj and not request.user.is_superuser:
            return self.readonly_fields + (
                "is_staff", "is_superuser", "user_permissions"
            )
        return self.readonly_fields


@admin.register(State)
class StateAdmin(admin.ModelAdmin):
    list_select_related = True
    list_display = ('name', 'priority', '_user_count')

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.annotate(user_count=Count("userprofile__id"))

    @admin.display(description="Users", ordering="user_count")
    def _user_count(self, obj):
        return obj.user_count

    fieldsets = (
        (None, {
            'fields': ('name', 'permissions', 'priority'),
        }),
        ('Membership', {
            'fields': (
                'public',
                'member_characters',
                'member_corporations',
                'member_alliances',
                'member_factions'
            ),
        })
    )
    filter_horizontal = [
        'member_characters',
        'member_corporations',
        'member_alliances',
        'member_factions',
        'permissions'
    ]

    def formfield_for_manytomany(self, db_field, request, **kwargs):
        """overriding this formfield to have sorted lists in the form"""
        if db_field.name == "member_characters":
            kwargs["queryset"] = EveCharacter.objects.all()\
                .order_by(Lower('character_name'))
        elif db_field.name == "member_corporations":
            kwargs["queryset"] = EveCorporationInfo.objects.all()\
                .order_by(Lower('corporation_name'))
        elif db_field.name == "member_alliances":
            kwargs["queryset"] = EveAllianceInfo.objects.all()\
                .order_by(Lower('alliance_name'))
        elif db_field.name == "member_factions":
            kwargs["queryset"] = EveFactionInfo.objects.all()\
                .order_by(Lower('faction_name'))
        elif db_field.name == "permissions":
            kwargs["queryset"] = Permission.objects.select_related("content_type").all()
        return super().formfield_for_manytomany(db_field, request, **kwargs)

    def has_delete_permission(self, request, obj=None):
        if obj == get_guest_state():
            return False
        return super().has_delete_permission(request, obj=obj)

    def get_fieldsets(self, request, obj=None):
        if obj == get_guest_state():
            return (
                (None, {
                    'fields': ('permissions', 'priority'),
                }),
            )
        return super().get_fieldsets(request, obj=obj)

    def get_readonly_fields(self, request, obj=None):
        if not request.user.is_superuser:
            return self.readonly_fields + ("permissions",)
        return self.readonly_fields


class BaseOwnershipAdmin(admin.ModelAdmin):
    list_select_related = (
        'user__profile__state', 'user__profile__main_character', 'character')
    list_display = (
        user_profile_pic,
        user_username,
        user_main_organization,
        'character',
    )
    search_fields = (
        'user__username',
        'character__character_name',
        'character__corporation_name',
        'character__alliance_name',
        'character__faction_name'
    )
    list_filter = (
        MainCorporationsFilter,
        MainAllianceFilter,
    )

    class Media:
        css = {
            "all": ("allianceauth/authentication/css/admin.css",)
        }

    def get_readonly_fields(self, request, obj=None):
        if obj and obj.pk:
            return 'owner_hash', 'character'
        return tuple()


@admin.register(OwnershipRecord)
class OwnershipRecordAdmin(BaseOwnershipAdmin):
    list_display = BaseOwnershipAdmin.list_display + ('created',)


@admin.register(CharacterOwnership)
class CharacterOwnershipAdmin(BaseOwnershipAdmin):
    def has_add_permission(self, request):
        return False


class PermissionAdmin(admin.ModelAdmin):
    actions = None
    readonly_fields = [field.name for field in BasePermission._meta.fields]
    search_fields = ('codename', )
    list_display = ('admin_name', 'name', 'codename', 'content_type')
    list_filter = ('content_type__app_label',)

    @staticmethod
    def admin_name(obj):
        return str(obj)

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def has_module_permission(self, request):
        return True

    def has_change_permission(self, request, obj=None):
        # can see list but not edit it
        return not obj


# Hack to allow registration of django.contrib.auth models in our authentication app
class User(BaseUser):
    class Meta:
        proxy = True
        verbose_name = BaseUser._meta.verbose_name
        verbose_name_plural = BaseUser._meta.verbose_name_plural


class Permission(BasePermission):
    class Meta:
        proxy = True
        verbose_name = BasePermission._meta.verbose_name
        verbose_name_plural = BasePermission._meta.verbose_name_plural


try:
    admin.site.unregister(BaseUser)
finally:
    admin.site.register(User, UserAdmin)
    admin.site.register(Permission, PermissionAdmin)


@receiver(pre_save, sender=User)
def redirect_pre_save(sender, signal=None, *args, **kwargs):
    pre_save.send(BaseUser, *args, **kwargs)


@receiver(post_save, sender=User)
def redirect_post_save(sender, signal=None, *args, **kwargs):
    post_save.send(BaseUser, *args, **kwargs)


@receiver(pre_delete, sender=User)
def redirect_pre_delete(sender, signal=None, *args, **kwargs):
    pre_delete.send(BaseUser, *args, **kwargs)


@receiver(post_delete, sender=User)
def redirect_post_delete(sender, signal=None, *args, **kwargs):
    post_delete.send(BaseUser, *args, **kwargs)


@receiver(m2m_changed, sender=User.groups.through)
def redirect_m2m_changed_groups(sender, signal=None, *args, **kwargs):
    m2m_changed.send(BaseUser, *args, **kwargs)


@receiver(m2m_changed, sender=User.user_permissions.through)
def redirect_m2m_changed_permissions(sender, signal=None, *args, **kwargs):
    m2m_changed.send(BaseUser, *args, **kwargs)



"""
Django system checks for Alliance Auth
"""

from typing import List
from django import db
from django.core.checks import CheckMessage, Error, register, Warning
from allianceauth.utils.cache import get_redis_client
from django.utils import timezone
from packaging.version import InvalidVersion, Version as Pep440Version
from celery import current_app
from django.conf import settings
from sqlite3.dbapi2 import sqlite_version_info

"""
A = System Packages
B = Configuration
"""


@register()
def django_settings(app_configs, **kwargs) -> List[CheckMessage]:
    """
    Check that Django settings are correctly configured

    :param app_configs:
    :type app_configs:
    :param kwargs:
    :type kwargs:
    :return:
    :rtype:
    """

    errors: List[CheckMessage] = []

    # Check for SITE_URL
    if hasattr(settings, "SITE_URL"):
        # Check if SITE_URL is empty
        if settings.SITE_URL == "":
            errors.append(
                Error(
                    msg="'SITE_URL' is empty.",
                    hint="Make sure to set 'SITE_URL' to the URL of your Auth instance. (Without trailing slash)",
                    id="allianceauth.checks.B011",
                )
            )
        # Check if SITE_URL has a trailing slash
        elif settings.SITE_URL[-1] == "/":
            errors.append(
                Warning(
                    msg="'SITE_URL' has a trailing slash. This may lead to incorrect links being generated by Auth.",
                    hint="",
                    id="allianceauth.checks.B005",
                )
            )
    # SITE_URL not found
    else:
        errors.append(
            Error(
                msg="No 'SITE_URL' found is settings. This may lead to incorrect links being generated by Auth or Errors in 3rd party modules.",
                hint="",
                id="allianceauth.checks.B006",
            )
        )

    # Check for CSRF_TRUSTED_ORIGINS
    if hasattr(settings, "CSRF_TRUSTED_ORIGINS") and hasattr(settings, "SITE_URL"):
        # Check if SITE_URL is not in CSRF_TRUSTED_ORIGINS
        if settings.SITE_URL not in settings.CSRF_TRUSTED_ORIGINS:
            errors.append(
                Warning(
                    msg="'SITE_URL' not found in 'CSRF_TRUSTED_ORIGINS'. Auth may not load pages correctly until this is rectified.",
                    hint="",
                    id="allianceauth.checks.B007",
                )
            )
    # CSRF_TRUSTED_ORIGINS not found
    else:
        errors.append(
            Error(
                msg="No 'CSRF_TRUSTED_ORIGINS' found is settings, Auth may not load pages correctly until this is rectified",
                hint="",
                id="allianceauth.checks.B008",
            )
        )

    # Check for ESI_USER_CONTACT_EMAIL
    if hasattr(settings, "ESI_USER_CONTACT_EMAIL"):
        # Check if ESI_USER_CONTACT_EMAIL is empty
        if settings.ESI_USER_CONTACT_EMAIL == "":
            errors.append(
                Error(
                    msg="'ESI_USER_CONTACT_EMAIL' is empty. A valid email is required as maintainer contact for CCP.",
                    hint="",
                    id="allianceauth.checks.B009",
                )
            )
    # ESI_USER_CONTACT_EMAIL not found
    else:
        errors.append(
            Error(
                msg="No 'ESI_USER_CONTACT_EMAIL' found is settings. A valid email is required as maintainer contact for CCP.",
                hint="",
                id="allianceauth.checks.B010",
            )
        )

    return errors


@register()
def system_package_redis(app_configs, **kwargs) -> List[CheckMessage]:
    """
    Check that Redis is a supported version

    :param app_configs:
    :type app_configs:
    :param kwargs:
    :type kwargs:
    :return:
    :rtype:
    """

    allianceauth_redis_install_link = "https://allianceauth.readthedocs.io/en/latest/installation/allianceauth.html#redis-and-other-tools"

    errors: List[CheckMessage] = []

    try:
        redis_version = Pep440Version(get_redis_client().info()["redis_version"])
    except InvalidVersion:
        errors.append(Warning("Unable to confirm Redis Version"))

        return errors

    if (
        redis_version.major == 7
        and redis_version.minor == 2
        and timezone.now()
        > timezone.datetime(year=2025, month=8, day=31, tzinfo=timezone.utc)
    ):
        errors.append(
            Error(
                msg=f"Redis {redis_version.public} in Security Support only, Updating Suggested",
                hint=allianceauth_redis_install_link,
                id="allianceauth.checks.A001",
            )
        )
    elif redis_version.major == 7 and redis_version.minor == 0:
        errors.append(
            Warning(
                msg=f"Redis {redis_version.public} in Security Support only, Updating Suggested",
                hint=allianceauth_redis_install_link,
                id="allianceauth.checks.A002",
            )
        )
    elif redis_version.major == 6 and redis_version.minor == 2:
        errors.append(
            Warning(
                msg=f"Redis {redis_version.public} in Security Support only, Updating Suggested",
                hint=allianceauth_redis_install_link,
                id="allianceauth.checks.A018",
            )
        )
    elif redis_version.major in [6, 5]:
        errors.append(
            Error(
                msg=f"Redis {redis_version.public} EOL",
                hint=allianceauth_redis_install_link,
                id="allianceauth.checks.A003",
            )
        )

    return errors


@register()
def system_package_mysql(app_configs, **kwargs) -> List[CheckMessage]:
    """
    Check that MySQL is a supported version

    :param app_configs:
    :type app_configs:
    :param kwargs:
    :type kwargs:
    :return:
    :rtype:
    """

    mysql_quick_guide_link = "https://dev.mysql.com/doc/mysql-apt-repo-quick-guide/en/"

    errors: List[CheckMessage] = []

    for connection in db.connections.all():
        if connection.vendor == "mysql":
            try:
                mysql_version = Pep440Version(
                    ".".join(str(i) for i in connection.mysql_version)
                )
            except InvalidVersion:
                errors.append(Warning("Unable to confirm MySQL Version"))

                return errors

            #  MySQL 8
            if mysql_version.major == 8:
                if mysql_version.minor == 4 and timezone.now() > timezone.datetime(
                    year=2032, month=4, day=30, tzinfo=timezone.utc
                ):
                    errors.append(
                        Error(
                            msg=f"MySQL {mysql_version.public} EOL",
                            hint=mysql_quick_guide_link,
                            id="allianceauth.checks.A004",
                        )
                    )
                elif mysql_version.minor == 3:
                    errors.append(
                        Warning(
                            msg=f"MySQL {mysql_version.public} Non LTS",
                            hint=mysql_quick_guide_link,
                            id="allianceauth.checks.A005",
                        )
                    )
                elif mysql_version.minor == 2:
                    errors.append(
                        Warning(
                            msg=f"MySQL {mysql_version.public} Non LTS",
                            hint=mysql_quick_guide_link,
                            id="allianceauth.checks.A006",
                        )
                    )
                elif mysql_version.minor == 1:
                    errors.append(
                        Error(
                            msg=f"MySQL {mysql_version.public} EOL",
                            hint=mysql_quick_guide_link,
                            id="allianceauth.checks.A007",
                        )
                    )
                elif mysql_version.minor == 0 and timezone.now() > timezone.datetime(
                    year=2026, month=4, day=30, tzinfo=timezone.utc
                ):
                    errors.append(
                        Error(
                            msg=f"MySQL {mysql_version.public} EOL",
                            hint=mysql_quick_guide_link,
                            id="allianceauth.checks.A008",
                        )
                    )

            # MySQL below 8
            # This will also catch Mariadb 5.x
            elif mysql_version.major < 8:
                errors.append(
                    Error(
                        msg=f"MySQL or MariaDB {mysql_version.public} EOL",
                        hint=mysql_quick_guide_link,
                        id="allianceauth.checks.A009",
                    )
                )

    return errors


@register()
def system_package_mariadb(app_configs, **kwargs) -> List[CheckMessage]:
    """
    Check that MariaDB is a supported version

    :param app_configs:
    :type app_configs:
    :param kwargs:
    :type kwargs:
    :return:
    :rtype:
    """

    mariadb_download_link = "https://mariadb.org/download/?t=repo-config"

    errors: List[CheckMessage] = []

    for connection in db.connections.all():
        # TODO: Find a way to determine MySQL vs. MariaDB
        if connection.vendor == "mysql":
            try:
                mariadb_version = Pep440Version(
                    ".".join(str(i) for i in connection.mysql_version)
                )
            except InvalidVersion:
                errors.append(Warning("Unable to confirm MariaDB Version"))

                return errors

            # MariaDB 11
            if mariadb_version.major == 11:
                if mariadb_version.minor == 4 and timezone.now() > timezone.datetime(
                    year=2029, month=5, day=19, tzinfo=timezone.utc
                ):
                    errors.append(
                        Error(
                            msg=f"MariaDB {mariadb_version.public} EOL",
                            hint=mariadb_download_link,
                            id="allianceauth.checks.A010",
                        )
                    )
                elif mariadb_version.minor == 2:
                    errors.append(
                        Warning(
                            msg=f"MariaDB {mariadb_version.public} Non LTS",
                            hint=mariadb_download_link,
                            id="allianceauth.checks.A018",
                        )
                    )

                    if timezone.now() > timezone.datetime(
                        year=2024, month=11, day=21, tzinfo=timezone.utc
                    ):
                        errors.append(
                            Error(
                                msg=f"MariaDB {mariadb_version.public} EOL",
                                hint=mariadb_download_link,
                                id="allianceauth.checks.A011",
                            )
                        )
                elif mariadb_version.minor == 1:
                    errors.append(
                        Warning(
                            msg=f"MariaDB {mariadb_version.public} Non LTS",
                            hint=mariadb_download_link,
                            id="allianceauth.checks.A019",
                        )
                    )

                    if timezone.now() > timezone.datetime(
                        year=2024, month=8, day=21, tzinfo=timezone.utc
                    ):
                        errors.append(
                            Error(
                                msg=f"MariaDB {mariadb_version.public} EOL",
                                hint=mariadb_download_link,
                                id="allianceauth.checks.A012",
                            )
                        )
                # Demote versions down here once EOL
                elif mariadb_version.minor in [0, 3]:
                    errors.append(
                        Error(
                            msg=f"MariaDB {mariadb_version.public} EOL",
                            hint=mariadb_download_link,
                            id="allianceauth.checks.A013",
                        )
                    )

            # MariaDB 10
            elif mariadb_version.major == 10:
                if mariadb_version.minor == 11 and timezone.now() > timezone.datetime(
                    year=2028, month=2, day=10, tzinfo=timezone.utc
                ):
                    errors.append(
                        Error(
                            msg=f"MariaDB {mariadb_version.public} EOL",
                            hint=mariadb_download_link,
                            id="allianceauth.checks.A014",
                        )
                    )
                elif mariadb_version.minor == 6 and timezone.now() > timezone.datetime(
                    year=2026, month=7, day=6, tzinfo=timezone.utc
                ):
                    errors.append(
                        Error(
                            msg=f"MariaDB {mariadb_version.public} EOL",
                            hint=mariadb_download_link,
                            id="allianceauth.checks.A0015",
                        )
                    )
                elif mariadb_version.minor == 5 and timezone.now() > timezone.datetime(
                    year=2025, month=6, day=24, tzinfo=timezone.utc
                ):
                    errors.append(
                        Error(
                            msg=f"MariaDB {mariadb_version.public} EOL",
                            hint=mariadb_download_link,
                            id="allianceauth.checks.A016",
                        )
                    )
                # Demote versions down here once EOL
                elif mariadb_version.minor in [0, 1, 2, 3, 4, 7, 9, 10]:
                    errors.append(
                        Error(
                            msg=f"MariaDB {mariadb_version.public} EOL",
                            hint=mariadb_download_link,
                            id="allianceauth.checks.A017",
                        )
                    )

    return errors


@register()
def system_package_sqlite(app_configs, **kwargs) -> List[CheckMessage]:
    """
    Check that SQLite is a supported version

    :param app_configs:
    :type app_configs:
    :param kwargs:
    :type kwargs:
    :return:
    :rtype:
    """

    errors: List[CheckMessage] = []

    for connection in db.connections.all():
        if connection.vendor == "sqlite":
            try:
                sqlite_version = Pep440Version(
                    ".".join(str(i) for i in sqlite_version_info)
                )
            except InvalidVersion:
                errors.append(Warning("Unable to confirm SQLite Version"))

                return errors
            if sqlite_version.major == 3 and sqlite_version.minor < 27:
                errors.append(
                    Error(
                        msg=f"SQLite {sqlite_version.public} Unsupported by Django",
                        hint="https://pkgs.org/download/sqlite3",
                        id="allianceauth.checks.A020",
                    )
                )

    return errors


@register()
def sql_settings(app_configs, **kwargs) -> List[CheckMessage]:
    """
    Check that SQL settings are correctly configured

    :param app_configs:
    :type app_configs:
    :param kwargs:
    :type kwargs:
    :return:
    :rtype:
    """

    errors: List[CheckMessage] = []

    for connection in db.connections.all():
        if connection.vendor == "mysql":
            try:
                if connection.settings_dict["OPTIONS"]["charset"] != "utf8mb4":
                    errors.append(
                        Error(
                            msg=f"SQL Charset is not set to utf8mb4 DB: {connection.alias}",
                            hint="https://gitlab.com/allianceauth/allianceauth/-/commit/89be2456fb2d741b86417e889da9b6129525bec8",
                            id="allianceauth.checks.B001",
                        )
                    )
            except KeyError:
                errors.append(
                    Error(
                        msg=f"SQL Charset is not set to utf8mb4 DB: {connection.alias}",
                        hint="https://gitlab.com/allianceauth/allianceauth/-/commit/89be2456fb2d741b86417e889da9b6129525bec8",
                        id="allianceauth.checks.B001",
                    )
                )

            # This hasn't actually been set on AA yet
            # try:
            #     if (
            #         connection.settings_dict["OPTIONS"]["collation"]
            #         != "utf8mb4_unicode_ci"
            #     ):
            #         errors.append(
            #             Error(
            #                 msg=f"SQL Collation is not set to utf8mb4_unicode_ci DB:{connection.alias}",
            #                 hint="https://gitlab.com/allianceauth/allianceauth/-/commit/89be2456fb2d741b86417e889da9b6129525bec8",
            #                 id="allianceauth.checks.B001",
            #             )
            #         )
            # except KeyError:
            #     errors.append(
            #         Error(
            #             msg=f"SQL Collation is not set to utf8mb4_unicode_ci DB:{connection.alias}",
            #             hint="https://gitlab.com/allianceauth/allianceauth/-/commit/89be2456fb2d741b86417e889da9b6129525bec8",
            #             id="allianceauth.checks.B001",
            #         )
            #     )

        # if connection.vendor == "sqlite":

    return errors


@register()
def celery_settings(app_configs, **kwargs) -> List[CheckMessage]:
    """
    Check that Celery settings are correctly configured

    :param app_configs:
    :type app_configs:
    :param kwargs:
    :type kwargs:
    :return:
    :rtype:
    """

    errors: List[CheckMessage] = []

    try:
        if current_app.conf.broker_transport_options != {
            "priority_steps": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            "queue_order_strategy": "priority",
        }:
            errors.append(
                Error(
                    msg="Celery Priorities are not set correctly",
                    hint="https://gitlab.com/allianceauth/allianceauth/-/commit/8861ec0a61790eca0261f1adc1cc04ca5f243cbc",
                    id="allianceauth.checks.B003",
                )
            )
    except KeyError:
        errors.append(
            Error(
                msg="Celery Priorities are not set",
                hint="https://gitlab.com/allianceauth/allianceauth/-/commit/8861ec0a61790eca0261f1adc1cc04ca5f243cbc",
                id="allianceauth.checks.B003",
            )
        )

    try:
        if not current_app.conf.broker_connection_retry_on_startup:
            errors.append(
                Error(
                    msg="Celery broker_connection_retry_on_startup not set correctly",
                    hint="https://gitlab.com/allianceauth/allianceauth/-/commit/380c41400b535447839e5552df2410af35a75280",
                    id="allianceauth.checks.B004",
                )
            )
    except KeyError:
        errors.append(
            Error(
                msg="Celery broker_connection_retry_on_startup not set",
                hint="https://gitlab.com/allianceauth/allianceauth/-/commit/380c41400b535447839e5552df2410af35a75280",
                id="allianceauth.checks.B004",
            )
        )

    return errors


# IDEAS

# Any other celery things weve manually changed over the years
# I'd be happy to add Community App checks, old versions the owners dont want to support etc.


# Check Default Collation on DB
# Check Charset Collation on all tables



from unittest.mock import patch

from django.test import TestCase
from allianceauth.tests.auth_utils import AuthUtils

from ..models import Notification


MODULE_PATH = 'allianceauth.notifications.models'


class TestUserNotify(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.create_user('magic_mike')
        AuthUtils.add_main_character(
            cls.user,
            'Magic Mike',
            '1',
            corp_id='2',
            corp_name='Pole Riders',
            corp_ticker='PRIDE',
            alliance_id='3',
            alliance_name='RIDERS'
        )

    @patch(MODULE_PATH + '.Notification.objects.invalidate_user_notification_cache')
    def test_save_will_invalidate_cache(self, mock_invalidate_user_notification_cache):
        obj = Notification.objects.notify_user(self.user, 'dummy')
        self.assertTrue(Notification.objects.filter(pk=obj.pk).exists())
        self.assertEqual(mock_invalidate_user_notification_cache.call_count, 1)

    @patch(MODULE_PATH + '.Notification.objects.invalidate_user_notification_cache')
    def test_delete_will_invalidate_cache(
        self, mock_invalidate_user_notification_cache
    ):
        obj = Notification.objects.notify_user(self.user, 'dummy')
        obj.delete()
        self.assertFalse(Notification.objects.filter(pk=obj.pk).exists())
        self.assertEqual(mock_invalidate_user_notification_cache.call_count, 2)

    def test_can_view(self):
        obj = Notification.objects.notify_user(self.user, 'dummy')
        self.assertFalse(obj.viewed)
        obj.mark_viewed()
        obj.refresh_from_db()
        self.assertTrue(obj.viewed)

    def test_can_set_level(self):
        obj = Notification.objects.notify_user(self.user, 'dummy', level='info')
        obj.set_level('ERROR')
        obj.refresh_from_db()
        self.assertEqual(obj.level, 'danger')

        obj.set_level('CRITICAL')
        obj.refresh_from_db()
        self.assertEqual(obj.level, 'danger')

        obj.set_level('WARN')
        obj.refresh_from_db()
        self.assertEqual(obj.level, 'warning')

        obj.set_level('INFO')
        obj.refresh_from_db()
        self.assertEqual(obj.level, 'info')

        obj.set_level('DEBUG')
        obj.refresh_from_db()
        self.assertEqual(obj.level, 'success')

        with self.assertRaises(ValueError):
            obj.set_level('XXX')



from django.contrib import admin

from .models import AnalyticsIdentifier, AnalyticsTokens
from solo.admin import SingletonModelAdmin


@admin.register(AnalyticsIdentifier)
class AnalyticsIdentifierAdmin(SingletonModelAdmin):
    search_fields = ['identifier', ]
    list_display = ['identifier', ]


@admin.register(AnalyticsTokens)
class AnalyticsTokensAdmin(admin.ModelAdmin):
    search_fields = ['name', ]
    list_display = ['name', 'type', ]



import sys
from copy import copy
from pathlib import Path


class StartupCommand:
    """Information about the command this Django instance was started with."""

    def __init__(self) -> None:
        self._argv = copy(sys.argv)

    @property
    def argv(self) -> list:
        """Return raw list of command line arguments."""
        return self._argv

    @property
    def script_name(self) -> str:
        """Return the base script name."""
        path = Path(self._argv[0])
        return path.name

    @property
    def is_management_command(self) -> bool:
        return self.script_name == "manage.py"



from django.core.management.base import BaseCommand
from allianceauth.authentication.models import UserProfile


class Command(BaseCommand):
    help = 'Ensures all main characters have an active ownership'

    def handle(self, *args, **options):
        profiles = UserProfile.objects.filter(main_character__isnull=False).filter(
            main_character__character_ownership__isnull=True)
        if profiles.exists():
            for profile in profiles:
                self.stdout.write(self.style.ERROR(
                    '{} does not have an ownership. Resetting user {} main character.'.format(profile.main_character,
                                                                                                profile.user)))
                profile.main_character = None
                profile.save()
            self.stdout.write(self.style.WARNING(f'Reset {profiles.count()} main characters.'))
        else:
            self.stdout.write(self.style.SUCCESS('All main characters have active ownership.'))



import urllib
from unittest.mock import Mock, patch

from requests.exceptions import HTTPError

from django.contrib.auth.models import User

from allianceauth.tests.auth_utils import AuthUtils
from allianceauth.utils.testing import NoSocketsTestCase

from ..app_settings import DISCORD_APP_ID, DISCORD_APP_SECRET, DISCORD_CALLBACK_URL
from ..discord_client import (
    DISCORD_OAUTH_BASE_URL,
    DISCORD_OAUTH_TOKEN_URL,
    DiscordApiBackoff,
    DiscordClient,
    RolesSet,
)
from ..discord_client.tests.factories import (
    TEST_GUILD_ID,
    TEST_USER_ID,
    TEST_USER_NAME,
    create_role,
    create_user,
)
from ..models import DiscordUser
from ..utils import set_logger_to_file
from . import MODULE_PATH, TEST_MAIN_NAME

logger = set_logger_to_file(MODULE_PATH + '.managers', __file__)


@patch(MODULE_PATH + '.managers.DISCORD_GUILD_ID', TEST_GUILD_ID)
@patch(MODULE_PATH + '.managers.DiscordClient', spec=DiscordClient)
@patch(MODULE_PATH + '.managers.create_bot_client', spec=True)
@patch(
    MODULE_PATH + '.models.DiscordUser.objects._exchange_auth_code_for_token', spec=True
)
@patch(MODULE_PATH + '.managers.calculate_roles_for_user', spec=True)
@patch(MODULE_PATH + '.managers.user_formatted_nick', spec=True)
class TestAddUser(NoSocketsTestCase):

    def setUp(self):
        self.user = AuthUtils.create_user(TEST_USER_NAME)
        self.access_token = 'accesstoken'

    def test_can_create_user_no_roles_no_nick(
        self,
        mock_user_formatted_nick,
        mock_calculate_roles_for_user,
        mock_exchange_auth_code_for_token,
        mock_create_bot_client,
        mock_DiscordClient,
    ):
        # given
        discord_user = create_user(id=TEST_USER_ID)
        mock_user_formatted_nick.return_value = None
        mock_exchange_auth_code_for_token.return_value = self.access_token
        mock_DiscordClient.return_value.current_user.return_value = discord_user
        mock_calculate_roles_for_user.return_value = RolesSet([]), None
        mock_create_bot_client.return_value.add_guild_member.return_value = True
        # when
        result = DiscordUser.objects.add_user(self.user, authorization_code='abcdef')
        # then
        self.assertTrue(result)
        self.assertTrue(
            DiscordUser.objects.filter(user=self.user, uid=TEST_USER_ID).exists()
        )
        _, kwargs = mock_create_bot_client.return_value.add_guild_member.call_args
        self.assertEqual(kwargs['guild_id'], TEST_GUILD_ID)
        self.assertEqual(kwargs['user_id'], TEST_USER_ID)
        self.assertEqual(kwargs['access_token'], self.access_token)
        self.assertFalse(kwargs['role_ids'])
        self.assertIsNone(kwargs['nick'])

    def test_can_create_user_with_roles_no_nick(
        self,
        mock_user_formatted_nick,
        mock_calculate_roles_for_user,
        mock_exchange_auth_code_for_token,
        mock_create_bot_client,
        mock_DiscordClient,
    ):
        # given
        role_a = create_role(id=1)
        role_b = create_role(id=2)
        roles_calculated = RolesSet([role_a, role_b])
        discord_user = create_user(id=TEST_USER_ID)
        mock_user_formatted_nick.return_value = None
        mock_exchange_auth_code_for_token.return_value = self.access_token
        mock_DiscordClient.return_value.current_user.return_value = discord_user
        mock_calculate_roles_for_user.return_value = roles_calculated, None
        mock_create_bot_client.return_value.add_guild_member.return_value = True
        # when
        result = DiscordUser.objects.add_user(self.user, authorization_code='abcdef')
        # then
        self.assertTrue(result)
        self.assertTrue(
            DiscordUser.objects.filter(user=self.user, uid=TEST_USER_ID).exists()
        )
        _, kwargs = mock_create_bot_client.return_value.add_guild_member.call_args
        self.assertEqual(kwargs['guild_id'], TEST_GUILD_ID)
        self.assertEqual(kwargs['user_id'], TEST_USER_ID)
        self.assertEqual(kwargs['access_token'], self.access_token)
        self.assertSetEqual(set(kwargs['role_ids']), {1, 2})
        self.assertIsNone(kwargs['nick'])

    def test_can_activate_existing_user_with_roles_no_nick(
        self,
        mock_user_formatted_nick,
        mock_calculate_roles_for_user,
        mock_exchange_auth_code_for_token,
        mock_create_bot_client,
        mock_DiscordClient,
    ):
        # given
        role_a = create_role(id=1)
        role_b = create_role(id=2)
        roles_calculated = RolesSet([role_a, role_b])
        discord_user = create_user(id=TEST_USER_ID)
        mock_user_formatted_nick.return_value = None
        mock_exchange_auth_code_for_token.return_value = self.access_token
        mock_DiscordClient.return_value.current_user.return_value = discord_user
        mock_calculate_roles_for_user.return_value = roles_calculated, False
        mock_create_bot_client.return_value.add_guild_member.return_value = None
        mock_create_bot_client.return_value.modify_guild_member.return_value = True
        # when
        result = DiscordUser.objects.add_user(self.user, authorization_code='abcdef')
        # then
        self.assertTrue(result)
        self.assertTrue(
            DiscordUser.objects.filter(user=self.user, uid=TEST_USER_ID).exists()
        )
        _, kwargs = mock_create_bot_client.return_value.modify_guild_member.call_args
        self.assertEqual(kwargs['guild_id'], TEST_GUILD_ID)
        self.assertEqual(kwargs['user_id'], TEST_USER_ID)
        self.assertSetEqual(set(kwargs['role_ids']), {1, 2})
        self.assertIsNone(kwargs['nick'])

    @patch(MODULE_PATH + '.managers.DISCORD_SYNC_NAMES', True)
    def test_can_create_user_no_roles_with_nick(
        self,
        mock_user_formatted_nick,
        mock_calculate_roles_for_user,
        mock_exchange_auth_code_for_token,
        mock_create_bot_client,
        mock_DiscordClient,
    ):
        # given
        discord_user = create_user(id=TEST_USER_ID)
        mock_user_formatted_nick.return_value = TEST_MAIN_NAME
        mock_exchange_auth_code_for_token.return_value = self.access_token
        mock_DiscordClient.return_value.current_user.return_value = discord_user
        mock_calculate_roles_for_user.return_value = RolesSet([]), None
        mock_create_bot_client.return_value.add_guild_member.return_value = True
        # when
        result = DiscordUser.objects.add_user(self.user, authorization_code='abcdef')
        # then
        self.assertTrue(result)
        self.assertTrue(
            DiscordUser.objects.filter(user=self.user, uid=TEST_USER_ID).exists()
        )
        _, kwargs = mock_create_bot_client.return_value.add_guild_member.call_args
        self.assertEqual(kwargs['guild_id'], TEST_GUILD_ID)
        self.assertEqual(kwargs['user_id'], TEST_USER_ID)
        self.assertEqual(kwargs['access_token'], self.access_token)
        self.assertFalse(kwargs['role_ids'])
        self.assertEqual(kwargs['nick'], TEST_MAIN_NAME)

    @patch(MODULE_PATH + '.managers.DISCORD_SYNC_NAMES', True)
    def test_can_activate_existing_user_no_roles_with_nick(
        self,
        mock_user_formatted_nick,
        mock_calculate_roles_for_user,
        mock_exchange_auth_code_for_token,
        mock_create_bot_client,
        mock_DiscordClient,
    ):
        # given
        discord_user = create_user(id=TEST_USER_ID)
        mock_user_formatted_nick.return_value = TEST_MAIN_NAME
        mock_exchange_auth_code_for_token.return_value = self.access_token
        mock_DiscordClient.return_value.current_user.return_value = discord_user
        mock_calculate_roles_for_user.return_value = RolesSet([]), False
        mock_create_bot_client.return_value.add_guild_member.return_value = None
        mock_create_bot_client.return_value.modify_guild_member.return_value = True
        # when
        result = DiscordUser.objects.add_user(self.user, authorization_code='abcdef')
        # then
        self.assertTrue(result)
        self.assertTrue(
            DiscordUser.objects.filter(user=self.user, uid=TEST_USER_ID).exists()
        )
        _, kwargs = mock_create_bot_client.return_value.modify_guild_member.call_args
        self.assertEqual(kwargs['guild_id'], TEST_GUILD_ID)
        self.assertEqual(kwargs['user_id'], TEST_USER_ID)
        self.assertFalse(kwargs['role_ids'])
        self.assertEqual(kwargs['nick'], TEST_MAIN_NAME)

    @patch(MODULE_PATH + '.managers.DISCORD_SYNC_NAMES', False)
    def test_can_create_user_no_roles_and_without_nick_if_turned_off(
        self,
        mock_user_formatted_nick,
        mock_calculate_roles_for_user,
        mock_exchange_auth_code_for_token,
        mock_create_bot_client,
        mock_DiscordClient,
    ):
        # given
        discord_user = create_user(id=TEST_USER_ID)
        mock_user_formatted_nick.return_value = TEST_MAIN_NAME
        mock_exchange_auth_code_for_token.return_value = self.access_token
        mock_DiscordClient.return_value.current_user.return_value = discord_user
        mock_calculate_roles_for_user.return_value = RolesSet([]), None
        mock_create_bot_client.return_value.add_guild_member.return_value = True
        # when
        result = DiscordUser.objects.add_user(self.user, authorization_code='abcdef')
        # then
        self.assertTrue(result)
        self.assertTrue(
            DiscordUser.objects.filter(user=self.user, uid=TEST_USER_ID).exists()
        )
        _, kwargs = mock_create_bot_client.return_value.add_guild_member.call_args
        self.assertEqual(kwargs['guild_id'], TEST_GUILD_ID)
        self.assertEqual(kwargs['user_id'], TEST_USER_ID)
        self.assertEqual(kwargs['access_token'], self.access_token)
        self.assertFalse(kwargs['role_ids'])
        self.assertIsNone(kwargs['nick'])

    def test_can_activate_existing_guild_member(
        self,
        mock_user_formatted_nick,
        mock_calculate_roles_for_user,
        mock_exchange_auth_code_for_token,
        mock_create_bot_client,
        mock_DiscordClient,
    ):
        # given
        discord_user = create_user(id=TEST_USER_ID)
        roles_calculated = RolesSet([create_role()])
        mock_user_formatted_nick.return_value = None
        mock_exchange_auth_code_for_token.return_value = self.access_token
        mock_DiscordClient.return_value.current_user.return_value = discord_user
        mock_calculate_roles_for_user.return_value = roles_calculated, False
        mock_create_bot_client.return_value.add_guild_member.return_value = None
        mock_create_bot_client.return_value.modify_guild_member.return_value = True
        # when
        result = DiscordUser.objects.add_user(self.user, authorization_code='abcdef')
        # then
        self.assertTrue(result)
        self.assertTrue(
            DiscordUser.objects.filter(user=self.user, uid=TEST_USER_ID).exists()
        )
        self.assertTrue(mock_create_bot_client.return_value.modify_guild_member.called)

    def test_can_activate_existing_member_with_roles(
        self,
        mock_user_formatted_nick,
        mock_calculate_roles_for_user,
        mock_exchange_auth_code_for_token,
        mock_create_bot_client,
        mock_DiscordClient,
    ):
        # given
        discord_user = create_user(id=TEST_USER_ID)
        roles_calculated = RolesSet([create_role(id=1)])
        mock_user_formatted_nick.return_value = None
        mock_exchange_auth_code_for_token.return_value = self.access_token
        mock_DiscordClient.return_value.current_user.return_value = discord_user
        mock_calculate_roles_for_user.return_value = roles_calculated, False
        mock_create_bot_client.return_value.add_guild_member.return_value = None
        mock_create_bot_client.return_value.modify_guild_member.return_value = True
        # when
        result = DiscordUser.objects.add_user(self.user, authorization_code='abcdef')
        # then
        self.assertTrue(result)
        self.assertTrue(
            DiscordUser.objects.filter(user=self.user, uid=TEST_USER_ID).exists()
        )
        _, kwargs = mock_create_bot_client.return_value.modify_guild_member.call_args
        self.assertSetEqual(set(kwargs['role_ids']), {1})

    def test_can_activate_existing_guild_member_failure(
        self,
        mock_user_formatted_nick,
        mock_calculate_roles_for_user,
        mock_exchange_auth_code_for_token,
        mock_create_bot_client,
        mock_DiscordClient,
    ):
        # given
        discord_user = create_user(id=TEST_USER_ID)
        roles_calculated = RolesSet([create_role()])
        mock_user_formatted_nick.return_value = None
        mock_exchange_auth_code_for_token.return_value = self.access_token
        mock_DiscordClient.return_value.current_user.return_value = discord_user
        mock_calculate_roles_for_user.return_value = roles_calculated, False
        mock_create_bot_client.return_value.add_guild_member.return_value = None
        mock_create_bot_client.return_value.modify_guild_member.return_value = False
        # when
        result = DiscordUser.objects.add_user(self.user, authorization_code='abcdef')
        # then
        self.assertFalse(result)
        self.assertFalse(
            DiscordUser.objects.filter(user=self.user, uid=TEST_USER_ID).exists()
        )
        self.assertTrue(mock_create_bot_client.return_value.modify_guild_member.called)

    def test_return_false_when_user_creation_fails(
        self,
        mock_user_formatted_nick,
        mock_calculate_roles_for_user,
        mock_exchange_auth_code_for_token,
        mock_create_bot_client,
        mock_DiscordClient,
    ):
        # given
        discord_user = create_user(id=TEST_USER_ID)
        mock_user_formatted_nick.return_value = None
        mock_exchange_auth_code_for_token.return_value = self.access_token
        mock_DiscordClient.return_value.current_user.return_value = discord_user
        mock_calculate_roles_for_user.return_value = RolesSet([]), None
        mock_create_bot_client.return_value.add_guild_member.return_value = False
        # when
        result = DiscordUser.objects.add_user(self.user, authorization_code='abcdef')
        # then
        self.assertFalse(result)
        self.assertFalse(
            DiscordUser.objects.filter(user=self.user, uid=TEST_USER_ID).exists()
        )
        self.assertTrue(mock_create_bot_client.return_value.add_guild_member.called)

    def test_return_false_when_on_api_backoff(
        self,
        mock_user_formatted_nick,
        mock_calculate_roles_for_user,
        mock_exchange_auth_code_for_token,
        mock_create_bot_client,
        mock_DiscordClient,
    ):
        # given
        discord_user = create_user(id=TEST_USER_ID)
        mock_user_formatted_nick.return_value = None
        mock_exchange_auth_code_for_token.return_value = self.access_token
        mock_DiscordClient.return_value.current_user.return_value = discord_user
        mock_calculate_roles_for_user.return_value = RolesSet([]), None
        mock_create_bot_client.return_value.add_guild_member.side_effect = \
            DiscordApiBackoff(999)
        # when
        result = DiscordUser.objects.add_user(self.user, authorization_code='abcdef')
        # then
        self.assertFalse(result)
        self.assertFalse(
            DiscordUser.objects.filter(user=self.user, uid=TEST_USER_ID).exists()
        )
        self.assertTrue(mock_create_bot_client.return_value.add_guild_member.called)

    def test_return_false_on_http_error(
        self,
        mock_user_formatted_nick,
        mock_calculate_roles_for_user,
        mock_exchange_auth_code_for_token,
        mock_create_bot_client,
        mock_DiscordClient,
    ):
        # given
        discord_user = create_user(id=TEST_USER_ID)
        mock_user_formatted_nick.return_value = None
        mock_exchange_auth_code_for_token.return_value = self.access_token
        mock_DiscordClient.return_value.current_user.return_value = discord_user
        mock_calculate_roles_for_user.return_value = RolesSet([]), None
        mock_exception = HTTPError('error')
        mock_exception.response = Mock()
        mock_exception.response.status_code = 500
        mock_create_bot_client.return_value.add_guild_member.side_effect = mock_exception
        # when
        result = DiscordUser.objects.add_user(self.user, authorization_code='abcdef')
        # then
        self.assertFalse(result)
        self.assertFalse(
            DiscordUser.objects.filter(user=self.user, uid=TEST_USER_ID).exists()
        )
        self.assertTrue(mock_create_bot_client.return_value.add_guild_member.called)


class TestOauthHelpers(NoSocketsTestCase):

    @patch(MODULE_PATH + '.managers.DISCORD_APP_ID', '123456')
    def test_generate_bot_add_url(self):
        bot_add_url = DiscordUser.objects.generate_bot_add_url()

        auth_url = DISCORD_OAUTH_BASE_URL
        real_bot_add_url = (
            f'{auth_url}?client_id=123456&scope=bot'
            f'&permissions={DiscordUser.objects.BOT_PERMISSIONS}'
        )
        self.assertEqual(bot_add_url, real_bot_add_url)

    def test_generate_oauth_redirect_url(self):
        oauth_url = DiscordUser.objects.generate_oauth_redirect_url()

        self.assertIn(DISCORD_OAUTH_BASE_URL, oauth_url)
        self.assertIn('+'.join(DiscordUser.objects.SCOPES), oauth_url)
        self.assertIn(DISCORD_APP_ID, oauth_url)
        self.assertIn(urllib.parse.quote_plus(DISCORD_CALLBACK_URL), oauth_url)

    @patch(MODULE_PATH + '.managers.OAuth2Session', spec=True)
    def test_process_callback_code(self, oauth):
        instance = oauth.return_value
        instance.fetch_token.return_value = {'access_token': 'mywonderfultoken'}

        token = DiscordUser.objects._exchange_auth_code_for_token('12345')

        self.assertTrue(oauth.called)
        args, kwargs = oauth.call_args
        self.assertEqual(args[0], DISCORD_APP_ID)
        self.assertEqual(kwargs['redirect_uri'], DISCORD_CALLBACK_URL)
        self.assertTrue(instance.fetch_token.called)
        args, kwargs = instance.fetch_token.call_args
        self.assertEqual(args[0], DISCORD_OAUTH_TOKEN_URL)
        self.assertEqual(kwargs['client_secret'], DISCORD_APP_SECRET)
        self.assertEqual(kwargs['code'], '12345')
        self.assertEqual(token, 'mywonderfultoken')


class TestUserHasAccount(NoSocketsTestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = AuthUtils.create_user(TEST_USER_NAME)

    def test_return_true_if_user_has_account(self):
        DiscordUser.objects.create(user=self.user, uid=TEST_USER_ID)
        self.assertTrue(DiscordUser.objects.user_has_account(self.user))

    def test_return_false_if_user_has_no_account(self):
        self.assertFalse(DiscordUser.objects.user_has_account(self.user))

    def test_return_false_if_user_does_not_exist(self):
        my_user = User(username='Dummy')
        my_user.save()
        self.assertFalse(DiscordUser.objects.user_has_account(my_user))

    def test_return_false_if_not_called_with_user_object(self):
        self.assertFalse(DiscordUser.objects.user_has_account('abc'))


class TestOtherMethods(NoSocketsTestCase):
    @patch(MODULE_PATH + '.managers.core_group_to_role', spec=True)
    def test_should_call_group_to_role(self, mock_core_group_to_role):
        # given
        role = create_role(id=1, name="alpha", managed=False)
        mock_core_group_to_role.return_value = role
        # when
        result = DiscordUser.objects.group_to_role(Mock())
        # then
        self.assertEqual(result["id"], 1)
        self.assertEqual(result["name"], "alpha")
        self.assertEqual(result["managed"], False)

    @patch(MODULE_PATH + '.managers.core_server_name', spec=True)
    def test_should_call_server_name(self, mock_core_server_name):
        # when
        DiscordUser.objects.server_name()
        # then
        self.assertTrue(mock_core_server_name.called)



from django.contrib import admin

from .models import SmfUser
from ...admin import ServicesUserAdmin


@admin.register(SmfUser)
class SmfUserAdmin(ServicesUserAdmin):
    list_display = ServicesUserAdmin.list_display + ('username',)
    search_fields = ServicesUserAdmin.search_fields + ('username', )



from django.test import TestCase

from allianceauth.menu.core.menu_item_hooks import (
    MenuItemHookCustom,
    gather_params,
    generate_hash,
)
from allianceauth.menu.tests.factories import create_menu_item_hook_function


class TestGenerateHash(TestCase):
    def test_should_generate_same_hash(self):
        # given
        hook = create_menu_item_hook_function()

        # when
        result_1 = generate_hash(hook())
        result_2 = generate_hash(hook())

        # then
        self.assertIsInstance(result_1, str)
        self.assertEqual(result_1, result_2)

    def test_should_generate_different_hashes(self):
        # given
        hook_1 = create_menu_item_hook_function()
        hook_2 = create_menu_item_hook_function()

        # when
        result_1 = generate_hash(hook_1())
        result_2 = generate_hash(hook_2())

        # then
        self.assertNotEqual(result_1, result_2)


class TestExtractParams(TestCase):
    def test_should_return_params(self):
        # given
        hook = create_menu_item_hook_function(text="Alpha", order=42)

        # when
        result = gather_params(hook())

        # then
        self.assertEqual(result.text, "Alpha")
        self.assertEqual(result.order, 42)
        self.assertIsInstance(result.hash, str)


class TestMenuItemHookCustom(TestCase):
    def test_should_create_minimal(self):
        # when
        obj = MenuItemHookCustom(text="text", classes="classes", url_name="url_name")

        # then
        self.assertEqual(obj.text, "text")
        self.assertEqual(obj.classes, "classes")
        self.assertEqual(obj.url_name, "url_name")
        self.assertEqual(obj.url, "")
        self.assertIsNone(obj.is_folder)
        self.assertEqual(obj.html_id, "")
        self.assertListEqual(obj.children, [])



from allianceauth.menu.hooks import MenuItemHook
from allianceauth.services.hooks import UrlHook

from allianceauth import hooks
from allianceauth.timerboard.views import dashboard_timers
from . import urls


class TimerboardMenu(MenuItemHook):
    def __init__(self):
        MenuItemHook.__init__(
            self, 'Structure Timers',
            'fa-regular fa-clock',
            'timerboard:view',
            navactive=['timerboard:']
        )

    def render(self, request):
        if request.user.has_perm('auth.timer_view'):
            return MenuItemHook.render(self, request)
        return ''


@hooks.register('menu_item_hook')
def register_menu():
    return TimerboardMenu()


@hooks.register('url_hook')
def register_url():
    return UrlHook(urls, 'timerboard', r'^timers/')


class NextTimersHook(hooks.DashboardItemHook):
    def __init__(self):  #TODO add the view permms so if they cant see it is not rendered
        hooks.DashboardItemHook.__init__(
            self,
            dashboard_timers,
            7
        )


@hooks.register('dashboard_hook')
def register_groups_hook():
    return NextTimersHook()



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class OpenfireServiceConfig(AppConfig):
    name = 'allianceauth.services.modules.openfire'
    label = 'openfire'
    verbose_name = _('Openfire Service')



from celery.schedules import crontab
import logging
from allianceauth.crontab.models import CronOffset
from django.db import ProgrammingError


logger = logging.getLogger(__name__)


def offset_cron(schedule: crontab) -> crontab:
    """Take a crontab and apply a series of precalculated offsets to spread out tasks execution on remote resources

    Args:
        schedule (crontab): celery.schedules.crontab()

    Returns:
        crontab: A crontab with offsetted Minute and Hour fields
    """

    try:
        cron_offset = CronOffset.get_solo()

        # Stops this shit from happening 0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23
        # It is only cosmetic, but still annoying
        if schedule._orig_minute == '*':
            new_minute = '*'
        else:
            new_minute = [(m + (round(60 * cron_offset.minute))) % 60 for m in schedule.minute]
        if schedule._orig_hour == '*':
            new_hour = '*'
        else:
            new_hour = [(m + (round(24 * cron_offset.hour))) % 24 for m in schedule.hour]

        return crontab(
            minute=",".join(str(m) for m in sorted(new_minute)),
            hour=",".join(str(h) for h in sorted(new_hour)),
            day_of_month=schedule._orig_day_of_month,
            month_of_year=schedule._orig_month_of_year,
            day_of_week=schedule._orig_day_of_week)

    except ProgrammingError as e:
        # If this is called before migrations are run hand back the default schedule
        # These offsets are stored in a Singleton Model,
        logger.error(e)
        return schedule

    except Exception as e:
        # We absolutely cant fail to hand back a schedule
        logger.error(e)
        return schedule



"""Signals for Task Statistics."""

from celery.signals import (
    task_failure, task_internal_error, task_retry, task_success, worker_ready,
)

from django.conf import settings

from .counters import failed_tasks, retried_tasks, succeeded_tasks


def reset_counters():
    """Reset all counters for the celery status."""
    succeeded_tasks.clear()
    failed_tasks.clear()
    retried_tasks.clear()


def is_enabled() -> bool:
    """Return True if task statistics are enabled, else return False."""
    return not bool(
        getattr(settings, "ALLIANCEAUTH_DASHBOARD_TASK_STATISTICS_DISABLED", False)
    )


@worker_ready.connect
def reset_counters_when_celery_restarted(*args, **kwargs):
    if is_enabled():
        reset_counters()


@task_success.connect
def record_task_succeeded(*args, **kwargs):
    if is_enabled():
        succeeded_tasks.add()


@task_retry.connect
def record_task_retried(*args, **kwargs):
    if is_enabled():
        retried_tasks.add()


@task_failure.connect
def record_task_failed(*args, **kwargs):
    if is_enabled():
        failed_tasks.add()


@task_internal_error.connect
def record_task_internal_error(*args, **kwargs):
    if is_enabled():
        failed_tasks.add()



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class PermissionsToolConfig(AppConfig):
    name = 'allianceauth.permissions_tool'
    label = 'permissions_tool'
    verbose_name = _('Permissions Audit')



from allianceauth import hooks
from allianceauth.theme.hooks import ThemeHook


class MateriaThemeHook(ThemeHook):
    """
    Bootswatch Materia Theme
    https://bootswatch.com/materia/
    """

    def __init__(self):
        ThemeHook.__init__(
            self,
            "Materia",
            "Material is the metaphor",
            html_tags={"data-theme": "materia"},
            css=[{
                "url": "https://cdnjs.cloudflare.com/ajax/libs/bootswatch/5.3.3/materia/bootstrap.min.css",
                "integrity": "sha512-2S9Do+uTmZmmJpdmAcOKdUrK/YslcvAuRfIF2ws8+BW9AvZXMRZM+o8Wq+PZrfISD6ZlIaeCWWZAdeprXIoYuQ=="
            }],
            js=[{
                "url": "https://cdnjs.cloudflare.com/ajax/libs/popper.js/2.11.8/umd/popper.min.js",
                "integrity": "sha512-TPh2Oxlg1zp+kz3nFA0C5vVC6leG/6mm1z9+mA81MI5eaUVqasPLO8Cuk4gMF4gUfP5etR73rgU/8PNMsSesoQ=="
            }, {
                "url": "https://cdnjs.cloudflare.com/ajax/libs/bootstrap/5.3.3/js/bootstrap.min.js",
                "integrity": "sha512-ykZ1QQr0Jy/4ZkvKuqWn4iF3lqPZyij9iRv6sGqLRdTPkY69YX6+7wvVGmsdBbiIfN/8OdsI7HABjvEok6ZopQ=="
            }],
            header_padding="5.25em"
        )


@hooks.register('theme_hook')
def register_materia_hook():
    return MateriaThemeHook()



# Generated by Django 3.2.10 on 2022-04-08 19:30

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('groupmanagement', '0018_reservedgroupname'),
    ]

    operations = [
        migrations.AddField(
            model_name='authgroup',
            name='restricted',
            field=models.BooleanField(default=False, help_text='Group is restricted. This means that adding or removing users for this group requires a superuser admin.'),
        ),
    ]



from django.test import TestCase
from allianceauth.tests.auth_utils import AuthUtils

from .. import notify
from ..models import Notification


class TestUserNotificationCount(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.create_user('magic_mike')
        AuthUtils.add_main_character(
            cls.user,
            'Magic Mike',
            '1',
            corp_id='2',
            corp_name='Pole Riders',
            corp_ticker='PRIDE',
            alliance_id='3',
            alliance_name='RIDERS'
        )

    def test_can_notify_short(self):
        # when
        notify(self.user, "dummy")
        # then
        self.assertEqual(Notification.objects.filter(user=self.user).count(), 1)

    def test_can_notify_full(self):
        # when
        notify(user=self.user, title="title", message="message", level="danger")
        # then
        obj = Notification.objects.first()
        self.assertEqual(obj.user, self.user)
        self.assertEqual(obj.title, "title")
        self.assertEqual(obj.message, "message")
        self.assertEqual(obj.level, "danger")



from django.urls import include
from django.contrib.auth.decorators import user_passes_test
from django.core.exceptions import PermissionDenied
from functools import wraps
from typing import Callable, Iterable, Optional

from django.urls import include
from django.contrib import messages
from django.contrib.auth.decorators import login_required, user_passes_test
from django.core.exceptions import PermissionDenied
from django.shortcuts import redirect
from django.utils.translation import gettext_lazy as _


def user_has_main_character(user):
    return bool(user.profile.main_character)


def decorate_url_patterns(
    urls, decorator: Callable, excluded_views: Optional[Iterable] = None
):
    """Decorate views given in url patterns except when they are explicitly excluded.

    Args:
        - urls: Django URL patterns
        - decorator: Decorator to be added to each view
        - exclude_views: Optional iterable of view names to be excluded
    """
    url_list, app_name, namespace = include(urls)

    def process_patterns(url_patterns):
        for pattern in url_patterns:
            if hasattr(pattern, 'url_patterns'):
                # this is an include - apply to all nested patterns
                process_patterns(pattern.url_patterns)
            else:
                # this is a pattern
                if excluded_views and pattern.lookup_str in excluded_views:
                    return
                pattern.callback = decorator(pattern.callback)

    process_patterns(url_list)
    return url_list, app_name, namespace


def main_character_required(view_func):
    @wraps(view_func)
    def _wrapped_view(request, *args, **kwargs):
        if user_has_main_character(request.user):
            return view_func(request, *args, **kwargs)

        messages.error(request, _('A main character is required to perform that action. Add one below.'))
        return redirect('authentication:dashboard')
    return login_required(_wrapped_view)


def permissions_required(perm, login_url=None, raise_exception=False):
    """
    Decorator for views that checks whether a user has a particular permission
    enabled, redirecting to the log-in page if necessary.
    If the raise_exception parameter is given the PermissionDenied exception
    is raised.

    This decorator is identical to the django permission_required except it
    allows for passing a tuple/list of perms that will return true if any one
    of them is present.
    """
    def check_perms(user):
        if isinstance(perm, str):
            perms = (perm,)
        else:
            perms = perm
        # First check if the user has the permission (even anon users)
        for perm_ in perms:
            perm_ = (perm_,)
            if user.has_perms(perm_):
                return True
        # In case the 403 handler should be called raise the exception
        if raise_exception:
            raise PermissionDenied
        # As the last resort, show the login form
        return False
    return user_passes_test(check_perms, login_url=login_url)



# Generated by Django 4.2.16 on 2024-10-07 13:39

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("mumble", "0012_mumble_client_info"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="mumbleuser",
            options={
                "permissions": (
                    ("access_mumble", "Can access the Mumble service"),
                    (
                        "view_connection_history",
                        "Can access the connection history of the Mumble service",
                    ),
                )
            },
        ),
        migrations.AlterField(
            model_name="mumbleuser",
            name="release",
            field=models.TextField(
                blank=True,
                editable=False,
                help_text="Client release. For official releases, this equals the version. For snapshots and git compiles, this will be something else.",
                max_length=254,
                null=True,
                verbose_name="Mumble Release",
            ),
        ),
    ]



"""Client for interacting with the Discord API."""

import json
import logging
from enum import IntEnum
from hashlib import md5
from http import HTTPStatus
from time import sleep
from typing import Iterable, List, Optional, Set, Tuple
from urllib.parse import urljoin
from uuid import uuid1

import requests
from requests.exceptions import HTTPError
from redis import Redis

from allianceauth.utils.cache import get_redis_client

from allianceauth import __title_useragent__, __url__, __version__

from .. import __title__
from ..utils import LoggerAddTag
from .app_settings import (
    DISCORD_API_BASE_URL,
    DISCORD_API_TIMEOUT_CONNECT,
    DISCORD_API_TIMEOUT_READ,
    DISCORD_DISABLE_ROLE_CREATION,
    DISCORD_GUILD_NAME_CACHE_MAX_AGE,
    DISCORD_ROLES_CACHE_MAX_AGE,
)
from .exceptions import DiscordRateLimitExhausted, DiscordTooManyRequestsError
from .helpers import RolesSet
from .models import Guild, GuildMember, Role, User

logger = LoggerAddTag(logging.getLogger(__name__), __title__)

# max requests that can be executed until reset
RATE_LIMIT_MAX_REQUESTS = 5

# Time until remaining requests are reset
RATE_LIMIT_RESETS_AFTER = 5000

# Delay used for API backoff in case no info returned from API on 429s
DEFAULT_BACKOFF_DELAY = 5000

# additional duration to compensate for potential clock discrepancies
# with the Discord server
DURATION_CONTINGENCY = 500

# Client will do a blocking wait rather than throwing a backoff exception if the
# time until next reset is below this threshold
WAIT_THRESHOLD = 250

# Minimum wait duration when doing a blocking wait
MINIMUM_BLOCKING_WAIT = 50

# If the rate limit resets soon we will wait it out and then retry to
# either get a remaining request from our cached counter
# or again wait out a short reset time and retry again.
# This could happen several times within a high concurrency situation,
# but must fail after x tries to avoid an infinite loop
RATE_LIMIT_RETRIES = 1000


class DiscordApiStatusCode(IntEnum):
    """Status code returned from the Discord API."""
    UNKNOWN_MEMBER = 10007  #:


class DiscordClient:
    """This class provides a web client for interacting with the Discord API.

    The client has rate limiting that supports concurrency.
    This means it is able to ensure the API rate limit is not violated,
    even when used concurrently, e.g. with multiple parallel celery tasks.

    In addition the client support proper API backoff.

    Synchronization of rate limit infos across multiple processes
    is implemented with Redis and thus requires Redis as Django cache backend.

    The cache is shared across all clients and processes (also using Redis).

    All durations are in milliseconds.

    Most errors from the API will raise a requests.HTTPError.

    Args:
        access_token: Discord access token used to authenticate all calls to the API
        redis: Redis instance to be used.
        is_rate_limited: Set to False to turn off rate limiting (use with care).
            If not specified will try to use the Redis instance
            from the default Django cache backend.

    Raises:
        ValueError: No access token provided
    """
    _KEY_GLOBAL_BACKOFF_UNTIL = 'DISCORD_GLOBAL_BACKOFF_UNTIL'
    _KEY_GLOBAL_RATE_LIMIT_REMAINING = 'DISCORD_GLOBAL_RATE_LIMIT_REMAINING'
    _KEYPREFIX_GUILD_NAME = 'DISCORD_GUILD_NAME'
    _KEYPREFIX_GUILD_ROLES = 'DISCORD_GUILD_ROLES'
    _KEYPREFIX_ROLE_NAME = 'DISCORD_ROLE_NAME'

    def __init__(
        self,
        access_token: str,
        redis: Redis = None,
        is_rate_limited: bool = True
    ) -> None:
        if not access_token:
            raise ValueError('You must provide an access token.')
        self._access_token = str(access_token)
        self._is_rate_limited = bool(is_rate_limited)
        if not redis:
            self._redis = get_redis_client()
            if not isinstance(self._redis, Redis):
                raise RuntimeError(
                    'This class requires a Redis client, but none was provided '
                    'and the default Django cache backend is not Redis either.'
                )
        else:
            self._redis = redis

        lua_1 = """
            if redis.call("exists", KEYS[1]) == 0 then
                redis.call("set", KEYS[1], ARGV[1], 'px', ARGV[2])
            end
            return redis.call("decr", KEYS[1])
        """
        self.__redis_script_decr_or_set = self._redis.register_script(lua_1)

        lua_2 = """
            local current_px = tonumber(redis.call("pttl", KEYS[1]))
            if current_px < tonumber(ARGV[2]) then
                return redis.call("set", KEYS[1], ARGV[1], 'px', ARGV[2])
            else
                return nil
            end
        """
        self.__redis_script_set_longer = self._redis.register_script(lua_2)

    @property
    def access_token(self) -> str:
        """Discord access token."""
        return self._access_token

    @property
    def is_rate_limited(self) -> bool:
        """Wether this instance is rate limited."""
        return self._is_rate_limited

    def __repr__(self):
        return f'{type(self).__name__}(access_token=...{self.access_token[-5:]})'

    def _redis_decr_or_set(self, name: str, value: str, px: int) -> bool:
        """Decrease the key value if it exists and returns the result else set the key.

        Implemented as Lua script to ensure atomicity.
        """
        return self.__redis_script_decr_or_set(
            keys=[str(name)], args=[str(value), int(px)]
        )

    def _redis_set_if_longer(self, name: str, value: str, px: int) -> bool:
        """Like set, but only goes through if either key doesn't exist
        or px would be extended.

        Implemented as Lua script to ensure atomicity.
        """
        return self.__redis_script_set_longer(
            keys=[str(name)], args=[str(value), int(px)]
        )

    # users

    def current_user(self) -> User:
        """Fetch user belonging to the current access_token."""
        authorization = f'Bearer {self.access_token}'
        r = self._api_request(
            method='get', route='users/@me', authorization=authorization
        )
        return User.from_dict(r.json())

    # guild

    def guild_infos(self, guild_id: int) -> Guild:
        """Fetch all basic infos about this guild.

        Args:
            guild_id: Discord ID of the guild
        """
        route = f"guilds/{guild_id}"
        r = self._api_request(method='get', route=route)
        return Guild.from_dict(r.json())

    def guild_name(self, guild_id: int, use_cache: bool = True) -> str:
        """Fetch the name of this guild (cached).

        Args:
            guild_id: Discord ID of the guild
            use_cache: When set to False will force an API call to get the server name

        Returns:
            Name of the server or an empty string if something went wrong.
        """
        key_name = self._guild_name_cache_key(guild_id)
        if use_cache:
            guild_name = self._redis_decode(self._redis.get(key_name))
        else:
            guild_name = ""
        if not guild_name:
            try:
                guild = self.guild_infos(guild_id)
            except HTTPError:
                guild_name = ""
            else:
                guild_name = guild.name
                self._redis.set(
                    name=key_name, value=guild_name, ex=DISCORD_GUILD_NAME_CACHE_MAX_AGE
                )
        return guild_name

    @classmethod
    def _guild_name_cache_key(cls, guild_id: int) -> str:
        """Construct key for accessing role given by name in the role cache.

        Args:
            guild_id: Discord ID of the guild
        """
        gen_key = DiscordClient._generate_hash(f'{guild_id}')
        return f'{cls._KEYPREFIX_GUILD_NAME}__{gen_key}'

    # guild roles

    def guild_roles(self, guild_id: int, use_cache: bool = True) -> Set[Role]:
        """Fetch all roles for this guild.

        Args:
            guild_id: Discord ID of the guild
            use_cache: If is set to False it will always hit the API to retrieve
                fresh data and update the cache.

        Returns:
        """
        cache_key = self._guild_roles_cache_key(guild_id)
        roles = None
        if use_cache:
            roles_raw = self._redis.get(name=cache_key)
            if roles_raw:
                logger.debug('Returning roles for guild %s from cache', guild_id)
                roles = json.loads(self._redis_decode(roles_raw))
            logger.debug('No roles for guild %s in cache', guild_id)
        if roles is None:
            route = f"guilds/{guild_id}/roles"
            r = self._api_request(method='get', route=route)
            roles = r.json()
            if not roles or not isinstance(roles, list):
                raise RuntimeError(
                    f"Unexpected response when fetching roles from API: {roles}"
                )
            self._redis.set(
                name=cache_key,
                value=json.dumps(roles),
                ex=DISCORD_ROLES_CACHE_MAX_AGE
            )
        return {Role.from_dict(role) for role in roles}

    def create_guild_role(
        self, guild_id: int, role_name: str, **kwargs
    ) -> Optional[Role]:
        """Create a new guild role with the given name.

        See official documentation for additional optional parameters.

        Note that Discord allows the creation of multiple roles with the same name,
        so to avoid duplicates it's important to check existing roles
        before creating new one

        Args:
            guild_id: Discord ID of the guild
            role_name: Name of new role to create

        Returns:
            new role on success
        """
        route = f"guilds/{guild_id}/roles"
        data = {'name': Role.sanitize_name(role_name)}
        data.update(kwargs)
        r = self._api_request(method='post', route=route, data=data)
        role = r.json()
        if role:
            self._invalidate_guild_roles_cache(guild_id)
            return Role.from_dict(role)
        return None

    def delete_guild_role(self, guild_id: int, role_id: int) -> bool:
        """Delete a guild role."""
        route = f"guilds/{guild_id}/roles/{role_id}"
        r = self._api_request(method='delete', route=route)
        if r.status_code == 204:
            self._invalidate_guild_roles_cache(guild_id)
            return True
        return False

    def _invalidate_guild_roles_cache(self, guild_id: int) -> None:
        cache_key = self._guild_roles_cache_key(guild_id)
        self._redis.delete(cache_key)
        logger.debug('Guild roles cache invalidated')

    @classmethod
    def _guild_roles_cache_key(cls, guild_id: int) -> str:
        """Construct key for accessing cached roles for a guild.

        Args:
            guild_id: Discord ID of the guild
        """
        gen_key = cls._generate_hash(f'{guild_id}')
        return f'{cls._KEYPREFIX_GUILD_ROLES}__{gen_key}'

    def match_role_from_name(self, guild_id: int, role_name: str) -> Optional[Role]:
        """Fetch Discord role matching the given name (cached).

        Args:
            guild_id: Discord ID of the guild
            role_name: Name of role

        Returns:
            Matching role or None if no match is found
        """
        guild_roles = RolesSet(self.guild_roles(guild_id))
        return guild_roles.role_by_name(role_name)

    def match_or_create_roles_from_names(
        self, guild_id: int, role_names: Iterable[str]
    ) -> List[Tuple[Role, bool]]:
        """Fetch or create Discord roles matching the given names (cached).

        Will try to match with existing roles names
        Non-existing roles will be created, then created flag will be True

        Args:
            guild_id: ID of guild
            role_names: list of name strings each defining a role

        Returns:
            List of tuple of Role and created flag
        """
        roles = list()
        guild_roles = RolesSet(self.guild_roles(guild_id))
        role_names_cleaned = {Role.sanitize_name(name) for name in role_names}
        for role_name in role_names_cleaned:
            role, created = self.match_or_create_role_from_name(
                guild_id=guild_id, role_name=role_name, guild_roles=guild_roles
            )
            if role:
                roles.append((role, created))
            if created:
                guild_roles = guild_roles.union(RolesSet([role]))
        return roles

    def match_or_create_role_from_name(
        self, guild_id: int, role_name: str, guild_roles: RolesSet = None
    ) -> Tuple[Role, bool]:
        """Fetch or create Discord role matching the given name.

        Will try to match with existing roles names
        Non-existing roles will be created, then created flag will be True

        Args:
            guild_id: ID of guild
            role_name: strings defining name of a role
            guild_roles: All known guild roles as RolesSet object.
                Helps to void redundant lookups of guild roles
                when this method is used multiple times.

        Returns:
            Tuple of Role and created flag
        """
        if not isinstance(role_name, str):
            raise TypeError('role_name must be of type string')

        created = False
        if guild_roles is None:
            guild_roles = RolesSet(self.guild_roles(guild_id))
        role = guild_roles.role_by_name(role_name)
        if not role:
            if not DISCORD_DISABLE_ROLE_CREATION:
                logger.debug('Need to create missing role: %s', role_name)
                role = self.create_guild_role(guild_id, role_name)
                created = True
            else:
                role = None
        return role, created

    def match_or_create_roles_from_names_2(
        self, guild_id: int, role_names: Iterable[str]
    ) -> RolesSet:
        """Fetch or create Discord role matching the given name.

        Wrapper for ``match_or_create_role_from_name()``

        Returns:
            Roles as RolesSet object.
        """
        return RolesSet.create_from_matched_roles(
            self.match_or_create_roles_from_names(
                guild_id=guild_id, role_names=role_names
            )
        )

    # guild members

    def add_guild_member(
        self,
        guild_id: int,
        user_id: int,
        access_token: str,
        role_ids: list = None,
        nick: str = None
    ) -> Optional[bool]:
        """Adds a user to the guild.

        Returns:
            - True when a new user was added
            - None if the user already existed
            - False when something went wrong or raises exception
        """
        route = f"guilds/{guild_id}/members/{user_id}"
        data = {
            'access_token': str(access_token)
        }
        if role_ids:
            data['roles'] = self._sanitize_role_ids(role_ids)
        if nick:
            data['nick'] = GuildMember.sanitize_nick(nick)
        r = self._api_request(method='put', route=route, data=data)
        r.raise_for_status()
        if r.status_code == 201:
            return True
        elif r.status_code == 204:
            return None
        return False

    def guild_member(self, guild_id: int, user_id: int) -> Optional[GuildMember]:
        """Fetch info for a guild member.

        Args:
            guild_id: Discord ID of the guild
            user_id: Discord ID of the user

        Returns:
            guild member or ``None`` if the user is not a member of the guild
        """
        route = f'guilds/{guild_id}/members/{user_id}'
        r = self._api_request(method='get', route=route, raise_for_status=False)
        if self._is_member_unknown_error(r):
            logger.warning("Discord user ID %s could not be found on server.", user_id)
            return None
        r.raise_for_status()
        return GuildMember.from_dict(r.json())

    def modify_guild_member(
        self, guild_id: int, user_id: int, role_ids: List[int] = None, nick: str = None
    ) -> Optional[bool]:
        """Set properties of a guild member.

        Args:
            guild_id: Discord ID of the guild
            user_id: Discord ID of the user
            roles_id: New list of role IDs (if provided)
            nick: New nickname (if provided)

        Returns
            - True when successful
            - None if user is not a member of this guild
            - False otherwise
        """
        if not role_ids and not nick:
            raise ValueError('Must specify role_ids or nick')

        if role_ids and not isinstance(role_ids, list):
            raise TypeError('role_ids must be a list type')

        data = dict()
        if role_ids:
            data['roles'] = self._sanitize_role_ids(role_ids)

        if nick:
            data['nick'] = GuildMember.sanitize_nick(nick)

        route = f"guilds/{guild_id}/members/{user_id}"
        r = self._api_request(
            method='patch', route=route, data=data, raise_for_status=False
        )
        if self._is_member_unknown_error(r):
            logger.warning('User ID %s is not a member of this guild', user_id)
            return None
        r.raise_for_status()
        if r.status_code == 204:
            return True
        return False

    def remove_guild_member(self, guild_id: int, user_id: int) -> Optional[bool]:
        """Remove a member from a guild.

        Args:
            guild_id: Discord ID of the guild
            user_id: Discord ID of the user

        Returns:
            - True when successful
            - None if member does not exist
            - False otherwise
        """
        route = f"guilds/{guild_id}/members/{user_id}"
        r = self._api_request(
            method='delete', route=route, raise_for_status=False
        )
        if self._is_member_unknown_error(r):
            logger.warning('User ID %s is not a member of this guild', user_id)
            return None
        r.raise_for_status()
        if r.status_code == 204:
            return True
        return False

    # Guild member roles

    def add_guild_member_role(
        self, guild_id: int, user_id: int, role_id: int
    ) -> Optional[bool]:
        """Adds a role to a guild member

        Returns:
        - True when successful
        - None if member does not exist
        - False otherwise
        """
        route = f"guilds/{guild_id}/members/{user_id}/roles/{role_id}"
        r = self._api_request(method='put', route=route, raise_for_status=False)
        if self._is_member_unknown_error(r):
            logger.warning('User ID %s is not a member of this guild', user_id)
            return None
        r.raise_for_status()
        if r.status_code == 204:
            return True
        return False

    def remove_guild_member_role(
        self, guild_id: int, user_id: int, role_id: int
    ) -> Optional[bool]:
        """Remove a role to a guild member

        Args:
            guild_id: Discord ID of the guild
            user_id: Discord ID of the user
            role_id: Discord ID of role to be removed

        Returns:
            - True when successful
            - None if member does not exist
            - False otherwise
        """
        route = f"guilds/{guild_id}/members/{user_id}/roles/{role_id}"
        r = self._api_request(method='delete', route=route, raise_for_status=False)
        if self._is_member_unknown_error(r):
            logger.warning('User ID %s is not a member of this guild', user_id)
            return None
        r.raise_for_status()
        if r.status_code == 204:
            return True
        return False

    def guild_member_roles(self, guild_id: int, user_id: int) -> Optional[RolesSet]:
        """Fetch the current guild roles of a guild member.

        Args:
        - guild_id: Discord guild ID
        - user_id: Discord user ID

        Returns:
        - Member roles
        - None if user is not a member of the guild
        """
        member_info = self.guild_member(guild_id=guild_id, user_id=user_id)
        if member_info is None:
            return None  # User is no longer a member
        guild_roles = RolesSet(self.guild_roles(guild_id=guild_id))
        logger.debug('Current guild roles: %s', guild_roles.ids())
        _roles = set(member_info.roles)
        if not guild_roles.has_roles(member_info.roles):
            guild_roles = RolesSet(
                self.guild_roles(guild_id=guild_id, use_cache=False)
            )
            if not guild_roles.has_roles(member_info.roles):
                role_ids = set(member_info.roles).difference(guild_roles.ids())
                logger.warning(f'Discord user {user_id} has unknown roles: {role_ids}')
                for _r in role_ids:
                    _roles.remove(_r)
        return guild_roles.subset(_roles)

    @classmethod
    def _is_member_unknown_error(cls, r: requests.Response) -> bool:
        try:
            result = (
                r.status_code == HTTPStatus.NOT_FOUND
                and r.json()['code'] == DiscordApiStatusCode.UNKNOWN_MEMBER
            )
        except (ValueError, KeyError):
            result = False

        return result

    # Internal methods

    def _api_request(
        self,
        method: str,
        route: str,
        data: dict = None,
        authorization: str = None,
        raise_for_status: bool = True
    ) -> requests.Response:
        """Core method for performing all API calls.

        Args:
            method: HTTP method of the request, e.g. "get"
            route: Route in the Discord API, e.g. "users/@me"
            data: Data to be send with the request
            authorization: The authorization string to be used.
                Will use the default bot token if not set.
            raise_for_status: Whether a requests exception is to be raised when not ok

        Returns:
            The raw response from the API
        """
        uid = uuid1().hex

        if not hasattr(requests, method):
            raise ValueError('Invalid method: %s' % method)

        if not authorization:
            authorization = f'Bot {self.access_token}'

        self._handle_ongoing_api_backoff(uid)
        if self.is_rate_limited:
            self._ensure_rate_limed_not_exhausted(uid)
        headers = {
            'User-Agent': f'{__title_useragent__}/{__version__} (+{__url__})',
            'accept': 'application/json',
            'X-RateLimit-Precision': 'millisecond',
            'authorization': str(authorization)
        }
        if data:
            headers['content-type'] = 'application/json'

        url = urljoin(DISCORD_API_BASE_URL, route)
        args = {
            'url': url,
            'headers': headers,
            'timeout': (DISCORD_API_TIMEOUT_CONNECT, DISCORD_API_TIMEOUT_READ)
        }
        if data:
            args['json'] = data

        logger.info('%s: sending %s request to url \'%s\'', uid, method.upper(), url)
        logger.debug('%s: request headers: %s', uid, headers)
        r = getattr(requests, method)(**args)
        logger.debug(
            '%s: returned status code %d with headers: %s',
            uid,
            r.status_code,
            r.headers
        )
        logger.debug('%s: response:\n%s', uid, r.text)
        if not r.ok:
            member_id = url.split("/")[-1]
            logger.error(
                '%s: Discord API returned error code %d for member ID %s with this response: %s.',
                uid,
                r.status_code,
                member_id,
                r.text
            )

        if r.status_code == HTTPStatus.TOO_MANY_REQUESTS:
            self._handle_new_api_backoff(r, uid)

        self._report_rate_limit_from_api(r, uid)

        if raise_for_status:
            r.raise_for_status()

        return r

    def _handle_ongoing_api_backoff(self, uid: str) -> None:
        """Check if api is currently on backoff.

        If on backoff: will do a blocking wait if it expires soon,
        else raises exception.
        """
        global_backoff_duration = self._redis.pttl(self._KEY_GLOBAL_BACKOFF_UNTIL)
        if global_backoff_duration > 0:
            if global_backoff_duration < WAIT_THRESHOLD:
                logger.info(
                    '%s: Global API backoff still ongoing for %s ms. Waiting.',
                    uid,
                    global_backoff_duration
                )
                sleep(global_backoff_duration / 1000)
            else:
                logger.info(
                    '%s: Global API backoff still ongoing for %s ms. Re-raising.',
                    uid,
                    global_backoff_duration
                )
                raise DiscordTooManyRequestsError(retry_after=global_backoff_duration)

    def _ensure_rate_limed_not_exhausted(self, uid: str) -> int:
        """Ensures that the rate limit is not exhausted.

        If exhausted: will do a blocking wait if rate limit resets soon,
        else raises exception

        returns requests remaining on success
        """
        for _ in range(RATE_LIMIT_RETRIES):
            requests_remaining = self._redis_decr_or_set(
                name=self._KEY_GLOBAL_RATE_LIMIT_REMAINING,
                value=RATE_LIMIT_MAX_REQUESTS,
                px=RATE_LIMIT_RESETS_AFTER + DURATION_CONTINGENCY
            )
            resets_in = max(
                MINIMUM_BLOCKING_WAIT,
                self._redis.pttl(self._KEY_GLOBAL_RATE_LIMIT_REMAINING)
            )
            if requests_remaining >= 0:
                logger.debug(
                    '%s: Got one of %d remaining requests until reset in %s ms',
                    uid,
                    requests_remaining + 1,
                    resets_in
                )
                return requests_remaining

            elif resets_in < WAIT_THRESHOLD:
                sleep(resets_in / 1000)
                logger.debug(
                    '%s: No requests remaining until reset in %d ms. '
                    'Waiting for reset.',
                    uid,
                    resets_in
                )
                continue

            else:
                logger.debug(
                    '%s: No requests remaining until reset in %d ms. '
                    'Raising exception.',
                    uid,
                    resets_in
                )
                raise DiscordRateLimitExhausted(resets_in)

        raise RuntimeError('Failed to handle rate limit after after too many tries.')

    def _handle_new_api_backoff(self, r: requests.Response, uid: str) -> None:
        """Raise exception for new API backoff error."""
        response = r.json()
        if 'retry_after' in response:
            try:
                retry_after = \
                    int(response['retry_after']) + DURATION_CONTINGENCY
            except ValueError:
                retry_after = DEFAULT_BACKOFF_DELAY
        else:
            retry_after = DEFAULT_BACKOFF_DELAY
        self._redis_set_if_longer(
            name=self._KEY_GLOBAL_BACKOFF_UNTIL,
            value='GLOBAL_API_BACKOFF',
            px=retry_after
        )
        logger.warning(
            "%s: Rate limit violated. Need to back off for at least %d ms",
            uid,
            retry_after
        )
        raise DiscordTooManyRequestsError(retry_after=retry_after)

    def _report_rate_limit_from_api(self, r, uid) -> None:
        """Try to log the current rate limit reported from API."""
        if (
            logger.getEffectiveLevel() <= logging.DEBUG
            and 'x-ratelimit-limit' in r.headers
            and 'x-ratelimit-remaining' in r.headers
            and 'x-ratelimit-reset-after' in r.headers
        ):
            try:
                limit = int(r.headers['x-ratelimit-limit'])
                remaining = int(r.headers['x-ratelimit-remaining'])
                reset_after = float(r.headers['x-ratelimit-reset-after']) * 1000
                if remaining + 1 == limit:
                    logger.debug(
                        '%s: Rate limit reported from API: %d requests per %s ms',
                        uid,
                        limit,
                        reset_after
                    )
            except ValueError:
                pass

    @staticmethod
    def _redis_decode(value: str) -> str:
        """Decode a string from Redis and passes through None and Booleans."""
        if value is not None and not isinstance(value, bool):
            return value.decode('utf-8')
        return value

    @staticmethod
    def _generate_hash(key: str) -> str:
        """Generate hash key for given string."""
        return md5(key.encode('utf-8')).hexdigest()

    @staticmethod
    def _sanitize_role_ids(role_ids: Iterable[int]) -> List[int]:
        """Sanitize a list of role IDs, i.e. make sure its a list of unique integers."""
        return [int(role_id) for role_id in set(role_ids)]



"""This script is for functional testing of the Discord client with a Discord server

It will run single requests of the various functions to validate
that they actually work - excluding those that require Oauth, or does not work
with a bot token. The results can be also seen in a special log file.

This script is design to be run manually as unit test, e.g. by running the following:

python manage.py test
allianceauth.services.modules.discord.discord_self.client.tests.piloting_functionality

To make it work please set the below mentioned environment variables for your server.
Since this may cause lots of 429s we'd recommend NOT to use your
alliance Discord server for this.
"""

from uuid import uuid1
import os
from unittest import TestCase
from time import sleep

from .. import DiscordClient
from ...utils import set_logger_to_file

logger = set_logger_to_file(
    'allianceauth.services.modules.discord.discord_self.client.client', __file__
)

# Make sure to set these environnement variables for your Discord server and user
DISCORD_GUILD_ID = os.environ['DISCORD_GUILD_ID']
DISCORD_BOT_TOKEN = os.environ['DISCORD_BOT_TOKEN']
DISCORD_USER_ID = os.environ['DISCORD_USER_ID']

RATE_LIMIT_DELAY_SECS = 1


class TestDiscordApiLive(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        logger.info('Live demo of the Discord API Client')
        cls.client = DiscordClient(DISCORD_BOT_TOKEN)

    def test_run_other_features(self):
        """runs features that have not been run in any of the other tests"""
        self.client.guild_infos(DISCORD_GUILD_ID)
        sleep(RATE_LIMIT_DELAY_SECS)

        self.client.guild_name(DISCORD_GUILD_ID)
        sleep(RATE_LIMIT_DELAY_SECS)

        self.client.match_or_create_role_from_name(DISCORD_GUILD_ID, 'Testrole')
        sleep(RATE_LIMIT_DELAY_SECS)

        self.client.match_or_create_roles_from_names(
            DISCORD_GUILD_ID, ['Testrole A', 'Testrole B']
        )
        sleep(RATE_LIMIT_DELAY_SECS)

    def test_create_and_remove_roles(self):
        # get base
        logger.info('guild_roles')
        expected = {role['id'] for role in self.client.guild_roles(DISCORD_GUILD_ID)}

        # add role
        role_name = 'my test role 12345678'
        logger.info('create_guild_role')
        new_role = self.client.create_guild_role(
            guild_id=DISCORD_GUILD_ID, role_name=role_name
        )
        sleep(RATE_LIMIT_DELAY_SECS)
        self.assertEqual(new_role['name'], role_name)

        # remove role again
        logger.info('delete_guild_role')
        self.client.delete_guild_role(
            guild_id=DISCORD_GUILD_ID, role_id=new_role['id']
        )
        sleep(RATE_LIMIT_DELAY_SECS)

        # verify it worked
        logger.info('guild_roles')
        role_ids = {role['id'] for role in self.client.guild_roles(DISCORD_GUILD_ID)}
        sleep(RATE_LIMIT_DELAY_SECS)
        self.assertSetEqual(role_ids, expected)

    def test_change_member_nick(self):
        # set new nick for user
        logger.info('modify_guild_member')
        new_nick = f'Testnick {uuid1().hex}'[:32]
        self.assertTrue(
            self.client.modify_guild_member(
                guild_id=DISCORD_GUILD_ID, user_id=DISCORD_USER_ID, nick=new_nick
            )
        )
        sleep(RATE_LIMIT_DELAY_SECS)

        # verify it is saved
        logger.info('guild_member')
        user = self.client.guild_member(DISCORD_GUILD_ID, DISCORD_USER_ID)
        sleep(RATE_LIMIT_DELAY_SECS)
        self.assertEqual(user['nick'], new_nick)

    def test_member_add_remove_roles(self):
        # create new guild role
        logger.info('create_guild_role')
        new_role = self.client.create_guild_role(
            guild_id=DISCORD_GUILD_ID, role_name='Special role 98765'
        )
        sleep(RATE_LIMIT_DELAY_SECS)
        new_role_id = new_role['id']

        # add to member
        logger.info('add_guild_member_role')
        self.assertTrue(
            self.client.add_guild_member_role(
                guild_id=DISCORD_GUILD_ID, user_id=DISCORD_USER_ID, role_id=new_role_id
            )
        )
        sleep(RATE_LIMIT_DELAY_SECS)

        # remove again
        logger.info('remove_guild_member_role')
        self.assertTrue(
            self.client.remove_guild_member_role(
                guild_id=DISCORD_GUILD_ID, user_id=DISCORD_USER_ID, role_id=new_role_id
            )
        )
        sleep(RATE_LIMIT_DELAY_SECS)



from django.urls import path

from . import views

app_name = 'optimer'

urlpatterns = [
    path('', views.optimer_view, name='view'),
    path('add/', views.add_optimer_view, name='add'),
    path('<int:optimer_id>/remove/', views.remove_optimer, name='remove'),
    path('<int:optimer_id>/edit/', views.edit_optimer, name='edit'),
]



import logging

from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _

logger = logging.getLogger(__name__)

# TODO discuss permissions for user defined links
# TODO define aa way for hooks to predefine a "parent" to create a sub menu from modules
# TODO Add user documentation


class MenuConfig(AppConfig):
    name = "allianceauth.menu"
    label = "menu"
    verbose_name = _("Menu")

    def ready(self):
        from allianceauth.menu.core import smart_sync

        smart_sync.reset_menu_items_sync()



from django.core.exceptions import ObjectDoesNotExist
from django_celery_beat.schedulers import (
    DatabaseScheduler
)
from django_celery_beat.models import CrontabSchedule
from django.db.utils import OperationalError, ProgrammingError

from celery import schedules
from celery.utils.log import get_logger

from allianceauth.crontab.models import CronOffset
from allianceauth.crontab.utils import offset_cron

logger = get_logger(__name__)


class OffsetDatabaseScheduler(DatabaseScheduler):
    """
    Customization of Django Celery Beat, Database Scheduler
    Takes the Celery Schedule from local.py and applies our AA Framework Cron Offset, if apply_offset is true
    Otherwise it passes it through as normal
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def update_from_dict(self, mapping):
        s = {}

        try:
            cron_offset = CronOffset.get_solo()
        except (OperationalError, ProgrammingError, ObjectDoesNotExist) as exc:
            # This is just incase we haven't migrated yet or something
            logger.warning(
                "OffsetDatabaseScheduler: Could not fetch CronOffset (%r). "
                "Defering to DatabaseScheduler",
                exc
            )
            return super().update_from_dict(mapping)

        for name, entry_fields in mapping.items():
            try:
                apply_offset = entry_fields.pop("apply_offset", False)  # Ensure this pops before django tries to save to ORM
                entry = self.Entry.from_entry(name, app=self.app, **entry_fields)

                if apply_offset:
                    entry_fields.update({"apply_offset": apply_offset})  # Reapply this as its gets pulled from config inconsistently.
                    schedule_obj = entry.schedule
                    if isinstance(schedule_obj, schedules.crontab):
                        offset_cs = CrontabSchedule.from_schedule(offset_cron(schedule_obj))
                        offset_cs, created = CrontabSchedule.objects.get_or_create(
                            minute=offset_cs.minute,
                            hour=offset_cs.hour,
                            day_of_month=offset_cs.day_of_month,
                            month_of_year=offset_cs.month_of_year,
                            day_of_week=offset_cs.day_of_week,
                            timezone=offset_cs.timezone,
                        )
                        entry.schedule = offset_cron(schedule_obj)  # This gets passed into Celery Beats Memory, important to keep it in sync with the model/DB
                        entry.model.crontab = offset_cs
                        entry.model.save()
                        logger.debug(f"Offset applied for '{name}' due to 'apply_offset' = True.")

                if entry.model.enabled:
                    s[name] = entry

            except Exception as e:
                logger.exception("Error updating schedule for %s: %r", name, e)

        self.schedule.update(s)



from unittest import TestCase
from unittest.mock import patch, MagicMock
from django.urls import URLPattern

from allianceauth.services.hooks import UrlHook
from allianceauth.urls import urls_from_apps

MODULE_PATH = "allianceauth.urls"


@patch(MODULE_PATH + ".main_character_required")
@patch(MODULE_PATH + ".decorate_url_patterns")
class TestUrlsFromApps(TestCase):
    def test_should_decorate_url_by_default(self, mock_decorate_url_patterns, mock_main_character_required):
        # given
        def hook_function():
            return UrlHook(urlconf_module, "my_namespace", r"^my_app/")

        view = MagicMock(name="view")
        path = MagicMock(spec=URLPattern, name="path")
        path.callback = view
        urlconf_module = [patch], "my_app"
        # when
        result = urls_from_apps([hook_function], [])
        # then
        self.assertIsInstance(result[0], URLPattern)
        self.assertTrue(mock_decorate_url_patterns.called)
        args, _ = mock_decorate_url_patterns.call_args
        decorator = args[1]
        self.assertEqual(decorator, mock_main_character_required)
        excluded_views = args[2]
        self.assertIsNone(excluded_views)

    def test_should_not_decorate_when_excluded(self, mock_decorate_url_patterns, mock_main_character_required):
        # given
        def hook_function():
            return UrlHook(urlconf_module, "my_namespace", r"^my_app/", ["excluded_view"])

        view = MagicMock(name="view")
        path = MagicMock(spec=URLPattern, name="path")
        path.callback = view
        urlconf_module = [patch], "my_app"
        # when
        result = urls_from_apps([hook_function], ["my_app"])
        # then
        self.assertIsInstance(result[0], URLPattern)
        self.assertTrue(mock_decorate_url_patterns.called)
        args, _ = mock_decorate_url_patterns.call_args
        decorator = args[1]
        self.assertEqual(decorator, mock_main_character_required)
        excluded_views = args[2]
        self.assertSetEqual(excluded_views, {"excluded_view"})

    def test_should_decorate_when_app_has_no_permission(self, mock_decorate_url_patterns, mock_main_character_required):
        # given
        def hook_function():
            return UrlHook(urlconf_module, "my_namespace", r"^my_app/", ["excluded_view"])

        view = MagicMock(name="view")
        path = MagicMock(spec=URLPattern, name="path")
        path.callback = view
        urlconf_module = [patch], "my_app"
        # when
        result = urls_from_apps([hook_function], ["other_app"])
        # then
        self.assertIsInstance(result[0], URLPattern)
        self.assertTrue(mock_decorate_url_patterns.called)
        args, _ = mock_decorate_url_patterns.call_args
        decorator = args[1]
        self.assertEqual(decorator, mock_main_character_required)
        excluded_views = args[2]
        self.assertIsNone(excluded_views)



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class MumbleServiceConfig(AppConfig):
    name = 'allianceauth.services.modules.mumble'
    label = 'mumble'
    verbose_name = _('Mumble Service')



from unittest import mock
from django.db.models.signals import pre_save, post_save, pre_delete, m2m_changed
from allianceauth.authentication.models import UserProfile
from allianceauth.authentication.signals import reassess_on_profile_save
from .. import signals
from ..models import AutogroupsConfig

MODULE_PATH = 'allianceauth.eveonline.autogroups'


def patch(target, *args, **kwargs):
    return mock.patch(f'{MODULE_PATH}{target}', *args, **kwargs)


def connect_signals():
    post_save.connect(receiver=reassess_on_profile_save, sender=UserProfile)
    pre_save.connect(receiver=signals.pre_save_config, sender=AutogroupsConfig)
    pre_delete.connect(receiver=signals.pre_delete_config, sender=AutogroupsConfig)
    post_save.connect(receiver=signals.check_groups_on_profile_update, sender=UserProfile)
    m2m_changed.connect(receiver=signals.autogroups_states_changed, sender=AutogroupsConfig.states.through)


def disconnect_signals():
    post_save.disconnect(receiver=reassess_on_profile_save, sender=UserProfile)
    pre_save.disconnect(receiver=signals.pre_save_config, sender=AutogroupsConfig)
    pre_delete.disconnect(receiver=signals.pre_delete_config, sender=AutogroupsConfig)
    post_save.disconnect(receiver=signals.check_groups_on_profile_update, sender=UserProfile)
    m2m_changed.disconnect(receiver=signals.autogroups_states_changed, sender=AutogroupsConfig.states.through)



from unittest.mock import patch
from django.test import TestCase

from allianceauth.crontab.models import CronOffset


class CronOffsetModelTest(TestCase):
    def test_cron_offset_is_singleton(self):
        """
        Test that CronOffset is indeed a singleton and that
        multiple calls to get_solo() return the same instance.
        """
        offset1 = CronOffset.get_solo()
        offset2 = CronOffset.get_solo()

        # They should be the exact same object in memory
        self.assertEqual(offset1.pk, offset2.pk)

    def test_default_values_random(self):
        """
        Test that the default values are set via random_default() when
        no explicit value is provided. We'll patch 'random.random' to
        produce predictable output.
        """
        with patch('allianceauth.crontab.models.random', return_value=0.1234):
            # Force creation of a new CronOffset by clearing the existing one
            CronOffset.objects.all().delete()

            offset = CronOffset.get_solo()  # This triggers creation

            # All fields should be 0.1234, because we patched random()
            self.assertAlmostEqual(offset.minute, 0.1234)
            self.assertAlmostEqual(offset.hour, 0.1234)
            self.assertAlmostEqual(offset.day_of_month, 0.1234)
            self.assertAlmostEqual(offset.month_of_year, 0.1234)
            self.assertAlmostEqual(offset.day_of_week, 0.1234)

    def test_update_offset_values(self):
        """
        Test that we can update the offsets and retrieve them.
        """
        offset = CronOffset.get_solo()
        offset.minute = 0.5
        offset.hour = 0.25
        offset.day_of_month = 0.75
        offset.month_of_year = 0.99
        offset.day_of_week = 0.33
        offset.save()

        # Retrieve again to ensure changes persist
        saved_offset = CronOffset.get_solo()
        self.assertEqual(saved_offset.minute, 0.5)
        self.assertEqual(saved_offset.hour, 0.25)
        self.assertEqual(saved_offset.day_of_month, 0.75)
        self.assertEqual(saved_offset.month_of_year, 0.99)
        self.assertEqual(saved_offset.day_of_week, 0.33)

    def test_str_representation(self):
        """
        Verify the __str__ method returns 'Cron Offsets'.
        """
        offset = CronOffset.get_solo()
        self.assertEqual(str(offset), "Cron Offsets")



from allianceauth.analytics.models import AnalyticsIdentifier

from django.test.testcases import TestCase

from uuid import uuid4


# Identifiers
uuid_1 = "ab33e241fbf042b6aa77c7655a768af7"
uuid_2 = "7aa6bd70701f44729af5e3095ff4b55c"


class TestAnalyticsIdentifier(TestCase):

    def test_identifier_random(self):
        self.assertNotEqual(AnalyticsIdentifier.get_solo(), uuid4)



# Generated by Django 1.10.1 on 2016-09-06 23:54

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('fleetactivitytracking', '0002_auto_20160905_2220'),
    ]

    operations = [
        migrations.AlterField(
            model_name='fatlink',
            name='fatdatetime',
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
    ]



from django.urls import include, path

app_name = 'example'

module_urls = [
    # Add your module URLs here
]

urlpatterns = [
    path('example/', include((module_urls, app_name), namespace=app_name)),
]



from django.urls import path

from . import views

app_name = 'srp'

urlpatterns = [
    # SRP URLS
    path('', views.srp_management, name='management'),
    path('all/', views.srp_management, {'all': True}, name='all'),
    path('<int:fleet_id>/view/', views.srp_fleet_view, name='fleet'),
    path('add/', views.srp_fleet_add_view, name='add'),
    path('<int:fleet_id>/edit/', views.srp_fleet_edit_view, name='edit'),
    path('<str:fleet_srp>/request/', views.srp_request_view, name='request'),

    # SRP URLS
    path('<int:fleet_id>/remove/', views.srp_fleet_remove, name='remove'),
    path('<int:fleet_id>/disable/', views.srp_fleet_disable, name='disable'),
    path('<int:fleet_id>/enable/', views.srp_fleet_enable, name='enable'),
    path('<int:fleet_id>/complete/', views.srp_fleet_mark_completed,
        name='mark_completed'),
    path('<int:fleet_id>/incomplete/', views.srp_fleet_mark_uncompleted,
        name='mark_uncompleted'),
    path('request/remove/', views.srp_request_remove,
        name="request_remove"),
    path('request/approve/', views.srp_request_approve,
        name='request_approve'),
    path('request/reject/', views.srp_request_reject,
        name='request_reject'),
    path('request/<int:fleet_srp_request_id>/update/', views.srp_request_update_amount,
        name="request_update_amount"),
]



from unittest import mock

from django.test import TestCase, RequestFactory
from django import urls
from django.contrib.auth.models import User, Group, Permission
from django.core.exceptions import ObjectDoesNotExist

from allianceauth.tests.auth_utils import AuthUtils

from .auth_hooks import MumbleService
from .models import MumbleUser
from .tasks import MumbleTasks

MODULE_PATH = 'allianceauth.services.modules.mumble'
DEFAULT_AUTH_GROUP = 'Member'


def add_permissions():
    permission = Permission.objects.get(codename='access_mumble')
    members = Group.objects.get_or_create(name=DEFAULT_AUTH_GROUP)[0]
    AuthUtils.add_permissions_to_groups([permission], [members])


class MumbleHooksTestCase(TestCase):
    def setUp(self):
        self.member = 'member_user'
        member = AuthUtils.create_member(self.member)
        AuthUtils.add_main_character(member, 'auth_member', '12345', corp_id='111', corp_name='Test Corporation', corp_ticker='TESTR')
        member = User.objects.get(pk=member.pk)
        MumbleUser.objects.create(user=member)
        self.none_user = 'none_user'
        none_user = AuthUtils.create_user(self.none_user)
        self.service = MumbleService
        add_permissions()

    def test_has_account(self):
        member = User.objects.get(username=self.member)
        none_user = User.objects.get(username=self.none_user)
        self.assertTrue(MumbleTasks.has_account(member))
        self.assertFalse(MumbleTasks.has_account(none_user))

    def test_service_enabled(self):
        service = self.service()
        member = User.objects.get(username=self.member)
        none_user = User.objects.get(username=self.none_user)

        self.assertTrue(service.service_active_for_user(member))
        self.assertFalse(service.service_active_for_user(none_user))

    @mock.patch(MODULE_PATH + '.tasks.User.mumble')
    def test_update_all_groups(self, mumble):
        service = self.service()
        service.update_all_groups()
        # Check member and blue user have groups updated
        self.assertTrue(mumble.update_groups.called)
        self.assertEqual(mumble.update_groups.call_count, 1)

    def test_update_groups(self):
        # Check member has Member group updated
        service = self.service()
        member = User.objects.get(username=self.member)
        member.mumble.groups = ''  # Remove the group set in setUp
        member.mumble.save()

        service.update_groups(member)

        mumble_user = MumbleUser.objects.get(user=member)
        self.assertIn(DEFAULT_AUTH_GROUP, mumble_user.groups)

        # Check none user does not have groups updated
        service = self.service()
        none_user = User.objects.get(username=self.none_user)
        result = service.update_groups(none_user)
        self.assertFalse(result)

    def test_validate_user(self):
        service = self.service()
        # Test member is not deleted
        member = User.objects.get(username=self.member)
        service.validate_user(member)
        self.assertTrue(member.mumble)

        # Test none user is deleted
        none_user = User.objects.get(username=self.none_user)
        MumbleUser.objects.create(user=none_user)
        service.validate_user(none_user)
        with self.assertRaises(ObjectDoesNotExist):
            none_mumble = User.objects.get(username=self.none_user).mumble

    def test_delete_user(self):
        member = User.objects.get(username=self.member)

        service = self.service()
        result = service.delete_user(member)

        self.assertTrue(result)
        with self.assertRaises(ObjectDoesNotExist):
            mumble_user = User.objects.get(username=self.member).mumble

    def test_render_services_ctrl(self):
        service = self.service()
        member = User.objects.get(username=self.member)
        request = RequestFactory().get('/services/')
        request.user = member

        response = service.render_services_ctrl(request)
        self.assertTemplateUsed(service.service_ctrl_template)
        self.assertIn(urls.reverse('mumble:deactivate'), response)
        self.assertIn(urls.reverse('mumble:reset_password'), response)
        self.assertIn(urls.reverse('mumble:set_password'), response)

        # Test register becomes available
        member.mumble.delete()
        member = User.objects.get(username=self.member)
        request.user = member
        response = service.render_services_ctrl(request)
        self.assertIn(urls.reverse('mumble:activate'), response)


class MumbleViewsTestCase(TestCase):
    def setUp(self):
        self.member = AuthUtils.create_member('auth_member')
        self.member.email = 'auth_member@example.com'
        self.member.save()
        AuthUtils.add_main_character(self.member, 'auth_member', '12345', corp_id='111', corp_name='Test Corporation', corp_ticker='TESTR')
        self.member = User.objects.get(pk=self.member.pk)
        add_permissions()

    def login(self):
        self.client.force_login(self.member)

    def test_activate_update(self):
        self.login()
        expected_username = 'auth_member'
        expected_displayname = '[TESTR]auth_member'
        response = self.client.get(urls.reverse('mumble:activate'), follow=False)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, expected_username)
        # create
        mumble_user = MumbleUser.objects.get(user=self.member)
        self.assertEqual(mumble_user.username, expected_username)
        self.assertTrue(MumbleUser.objects.user_exists(expected_username))
        self.assertEqual(str(mumble_user), expected_username)
        self.assertEqual(mumble_user.display_name, expected_displayname)
        self.assertTrue(mumble_user.pwhash)
        self.assertIn('Guest', mumble_user.groups)
        self.assertIn('Member', mumble_user.groups)
        self.assertIn(',', mumble_user.groups)
        # test update
        self.member.profile.main_character.character_name = "auth_member_updated"
        self.member.profile.main_character.corporation_ticker = "TESTU"
        self.member.profile.main_character.save()
        mumble_user.update_display_name()
        mumble_user = MumbleUser.objects.get(user=self.member)
        expected_displayname = '[TESTU]auth_member_updated'
        self.assertEqual(mumble_user.username, expected_username)
        self.assertTrue(MumbleUser.objects.user_exists(expected_username))
        self.assertEqual(str(mumble_user), expected_username)
        self.assertEqual(mumble_user.display_name, expected_displayname)
        self.assertTrue(mumble_user.pwhash)
        self.assertIn('Guest', mumble_user.groups)
        self.assertIn('Member', mumble_user.groups)
        self.assertIn(',', mumble_user.groups)


    def test_deactivate_post(self):
        self.login()
        MumbleUser.objects.create(user=self.member)

        response = self.client.post(urls.reverse('mumble:deactivate'))

        self.assertRedirects(response, expected_url=urls.reverse('services:services'), target_status_code=200)
        with self.assertRaises(ObjectDoesNotExist):
            mumble_user = User.objects.get(pk=self.member.pk).mumble

    def test_set_password(self):
        self.login()
        created = MumbleUser.objects.create(user=self.member)
        old_pwd = created.credentials.get('password')

        response = self.client.post(urls.reverse('mumble:set_password'), data={'password': '1234asdf'})

        self.assertNotEqual(MumbleUser.objects.get(user=self.member).pwhash, old_pwd)
        self.assertRedirects(response, expected_url=urls.reverse('services:services'), target_status_code=200)

    def test_reset_password(self):
        self.login()
        created = MumbleUser.objects.create(user=self.member)
        old_pwd = created.credentials.get('password')

        response = self.client.get(urls.reverse('mumble:reset_password'))

        self.assertNotEqual(MumbleUser.objects.get(user=self.member).pwhash, old_pwd)
        self.assertTemplateUsed(response, 'services/service_credentials.html')
        self.assertContains(response, 'auth_member')

class MumbleManagerTestCase(TestCase):
    def setUp(self):
        from .models import MumbleManager
        self.manager = MumbleManager

    def test_generate_random_password(self):
        password = self.manager.generate_random_pass()

        self.assertEqual(len(password), 16)
        self.assertIsInstance(password, str)

    def test_gen_pwhash(self):
        pwhash = self.manager.gen_pwhash('test')

        self.assertEqual(pwhash[:15], '$bcrypt-sha256$')
        self.assertEqual(len(pwhash), 83)



# Generated by Django 4.2 on 2024-09-13 09:46

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('authentication', '0023_alter_userprofile_language'),
    ]

    operations = [
        migrations.AlterField(
            model_name='userprofile',
            name='language',
            field=models.CharField(blank=True, choices=[('en', 'English'), ('cs-cz', 'Czech'), ('de', 'German'), ('es', 'Spanish'), ('it-it', 'Italian'), ('ja', 'Japanese'), ('ko-kr', 'Korean'), ('fr-fr', 'French'), ('ru', 'Russian'), ('nl-nl', 'Dutch'), ('pl-pl', 'Polish'), ('uk', 'Ukrainian'), ('zh-hans', 'Simplified Chinese')], default='', max_length=10, verbose_name='Language'),
        ),
    ]
